# $Id: icmp.bro,v 1.8 2004/09/24 21:37:45 vern Exp $

@load hot
@load weird
@load conn
@load scan

global icmp_file = open_log_file("icmp");

redef capture_filters += { ["icmp"] = "icmp" };

module ICMP;

export {

	redef enum Notice += {
		ICMPAsymPayload,	# payload in echo req-resp not the same
		ICMPConnectionPair,	# too many ICMPs between hosts

		# The following isn't presently sufficiently useful due
		# to cold start and packet drops.
		# ICMPUnpairedEchoReply,	# no EchoRequest seen for EchoReply
	};

	# Whether to log ICMP packets to conn.log.
	const log_as_connection = T &redef;

	# Whether to log detailed information icmp.log.
	const log_details = T &redef;

	# ICMP scan detection.
	global detect_scans = T &redef;
	const scan_threshold = 25 &redef;

	# Analysis of connection pairs.
	global detect_conn_pairs = F &redef;	# switch for connection pair
	global detect_payload_asym = F &redef;	# switch for echo payload
	const conn_pair_threshold = 200 &redef;
}

global distinct_pairs: set[addr, addr];
global num_distinct_pairs: table[addr] of count &default=0;

global conn_pair: table[addr, addr] of count &default=0;


type flow_id: record {
	orig_h: addr;
	resp_h: addr;
	id: count;
};

type flow_info: record {
	start_time: time;
	last_time: time;
	orig_bytes: count;
	resp_bytes: count;
	payload: string;
};

const names: table[count] of string = {
	[0] = "echo_reply",
	[3] = "unreach",
	[4] = "quench",
	[5] = "redirect",
	[8] = "echo_req",
	[9] = "router_adv",
	[10] = "router_sol",
	[11] = "time_xcd",
	[12] = "param_prob",
	[13] = "tstamp_req",
	[14] = "tstamp_reply",
	[15] = "info_req",
	[16] = "info_reply",
	[17] = "mask_req",
	[18] = "mask_reply",
} &default = function(n: count): string { return fmt("icmp-%d", n); };


global flows: table[flow_id] of flow_info &read_expire = 45 sec;

event icmp_sent(icmp: icmp_conn)
	{
	# This is for single, non-flow-oriented ICMP packets.
	if ( log_as_connection )
		{
		print conn_file, fmt("%.6f %.6f %s %s %s %s %s %s %s %s %s",
			network_time(), 0.0, icmp$orig_h, icmp$resp_h,
			names[icmp$itype], icmp$itype, icmp$icode, "icmp",
			icmp$len, "0", "SH");
		}
	else
		{
		print icmp_file, fmt("%.6f %.6f %s %s %s %s %s %s %s %s %s",
			network_time(), 0.0, icmp$orig_h, icmp$resp_h,
			names[icmp$itype], icmp$itype, icmp$icode, "icmp",
			icmp$len, "0", "SH");
		}
	}

event flow_summary(flow: flow_id, last_time: time)
	{
	if ( flow !in flows )
		return;

	local fi = flows[flow];

	if ( fi$last_time == last_time )
		{
		local local_init = is_local_addr(flow$orig_h);
		local local_addr = local_init ? flow$orig_h : flow$resp_h;
		local remote_addr = local_init ? flow$resp_h : flow$orig_h;
		local flags = local_init ? "L" : "";

		local state: string;
		if ( fi$orig_bytes > 0 )
			{
			if ( fi$resp_bytes > 0 )
				state = "SF";
			else
				state = "SH";
			}
		else if ( fi$resp_bytes > 0 )
			state = "SHR";
		else
			state = "OTH";

		if ( log_as_connection )
			{
			print conn_file, fmt("%.6f %.6f %s %s %s %s %s %s %s",
				fi$start_time, fi$last_time - fi$start_time, flow$orig_h,
				flow$resp_h, "icmp_echo", fi$orig_bytes, fi$resp_bytes,
				state, flags);
			}
		else
			{
			print icmp_file, fmt("%.6f %.6f %s %s %s %s %s %s %s",
				fi$start_time, fi$last_time - fi$start_time, flow$orig_h,
				flow$resp_h, "icmp_echo", fi$orig_bytes, fi$resp_bytes,
				state, flags);
			}

		delete flows[flow];
		}
	}

function update_flow(icmp: icmp_conn, id: count, is_orig: bool, payload: string)
	{
	local fid: flow_id;
	fid$orig_h = is_orig ? icmp$orig_h : icmp$resp_h;
	fid$resp_h = is_orig ? icmp$resp_h : icmp$orig_h;
	fid$id = id;

	if ( fid !in flows )
		{
		local info: flow_info;
		info$start_time = network_time();
		info$orig_bytes = info$resp_bytes = 0;
		info$payload = payload;	# checked in icmp_echo_reply
		flows[fid] = info;
		}

	local fi = flows[fid];

	fi$last_time = network_time();

	if ( is_orig )
		fi$orig_bytes = fi$orig_bytes + byte_len(payload);
	else
		fi$resp_bytes = fi$resp_bytes + byte_len(payload);

	schedule +30sec { flow_summary(fid, fi$last_time) };
	}

event icmp_echo_request(icmp: icmp_conn, id: count, seq: count, payload: string)
	{
	update_flow(icmp, id, T, payload);

	# Simple ping scan detector.
	if ( detect_scans && [icmp$orig_h, icmp$resp_h] !in distinct_pairs )
		{
		add distinct_pairs[icmp$orig_h, icmp$resp_h];

		if ( ++num_distinct_pairs[icmp$orig_h] == scan_threshold )
			{
			NOTICE([$note=AddressScan, $src=icmp$orig_h,
				$msg=fmt("%s has icmp echo scanned %s hosts",
				icmp$orig_h, scan_threshold)]);
			}
		}

	if ( detect_conn_pairs )
		{
		if ( ++conn_pair[icmp$orig_h, icmp$resp_h] == conn_pair_threshold )
			{
			NOTICE([$note=ICMPConnectionPair,
				$msg=fmt("ICMP connection threshold exceeded : %s -> %s",
				icmp$orig_h, icmp$resp_h)]);
			}
		}
	}

event icmp_echo_reply(icmp: icmp_conn, id: count, seq: count, payload: string)
	{
	# Check payload with the associated flow.

	local fid: flow_id;
	fid$orig_h = icmp$resp_h;	# we know the expected results since
	fid$resp_h = icmp$orig_h;	#  it's an echo reply
	fid$id = id;

	if ( fid !in flows )
		{
# 		NOTICE([$note=ICMPUnpairedEchoReply,
# 			$msg=fmt("ICMP echo reply w/o request: %s -> %s",
# 				icmp$orig_h, icmp$resp_h)]);
		}
	else
		{
		if ( detect_payload_asym )
			{
			local fi = flows[fid];
			local pl = fi$payload;

			if ( pl != payload )
				{
				NOTICE([$note=ICMPAsymPayload,
					$msg=fmt("ICMP payload inconsistancy: %s(%s) -> %s(%s)",
					icmp$orig_h, byte_len(fi$payload),
					icmp$resp_h, byte_len(payload))]);
				}
			}
		}

	update_flow(icmp, id, F, payload);
	}

event icmp_unreachable(icmp: icmp_conn, code: count, context: icmp_context)
	{
	if ( active_connection(context$id) )
		{
		local c = connection_record(context$id);
		local os = c$orig$state;
		local rs = c$resp$state;
		local is_attempt =
			is_tcp_port(c$id$orig_p) ?
				(os == TCP_SYN_SENT && rs == TCP_INACTIVE) :
				(os == UDP_ACTIVE && rs == UDP_INACTIVE);
		}

#	print icmp_file, fmt("%.6f %.6f %s %s %s %s %s %s %s",
#		fi$start_time, fi$last_time - fi$start_time, "icmp_echo",
#		fi$orig_bytes, fi$resp_bytes,
#		local_addr, remote_addr, state, flags);

	if ( log_details )
		### This should be structured to look like other icmp_file
		### entries (so, at a minimum, a timestamp, formatted to %.6f).
		print icmp_file, "unreach", icmp, code, context;

	if ( log_as_connection )
		{
		print conn_file, fmt("%.6f %.6f %s %s %s %s %s %s %s %s %s",
			network_time(), 0.0, icmp$orig_h, icmp$resp_h,
			names[icmp$itype], icmp$itype, icmp$icode, "icmp",
			icmp$len, "0", "SH");
		}
	}
