# $Id: notice.bro,v 1.4 2004/10/30 00:04:02 rwinslow Exp $

type Notice: enum {
	Unspecified,	# place-holder - shouldn't be used
};

type notice_info: record {
	note: Notice;
	msg: string;
	sub: string &optional;	# sub-message

	conn: connection &optional;	# connection associated with notice
	iconn: icmp_conn &optional;	# associated ICMP "connection"
	id: conn_id &optional;	# connection-ID, if we don't have a connection handy
	src: addr &optional;	# source address, if we don't have a connection
	dst: addr &optional;	# destination address

	p: port &optional;	# associated port, if we don't have a conn.

	# The following are detailed attributes that are associated with some
	# notices, but not all.

	user: string &optional;

	filename: string &optional;

	method: string &optional;
	URL: string &optional;

	n: count &optional;	# associated count, or perhaps status code

	# Source that raised this notice (automatically set).
	src_peer: event_peer &optional;
	
	# Reverse connection direction flag.  Used for things like signatures
	# where the attack originator can be different from the connection
	# originator.
	invert_conn: bool &optional;
};

type NoticeAction: enum {
	# Similar to WeirdAction in weird.bro.
	NOTICE_IGNORE, NOTICE_FILE,
	NOTICE_ALARM_ALWAYS, NOTICE_ALARM_PER_CONN,
	NOTICE_ALARM_PER_ORIG, NOTICE_ALARM_ONCE,
};

type notice_policy_item: record {
	result: NoticeAction;
	pred: function(n: notice_info): bool;
	priority: count;
};

global notice_policy: set[notice_policy_item] = {
	[$pred(n: notice_info) = { return T; },
	 $result = NOTICE_ALARM_ALWAYS,
	 $priority = 0],
} &redef;

# Table that maps notices into a function that should be called
# to determine the action.
global notice_action_filters:
	table[Notice] of function(n: notice_info): NoticeAction &redef;

function ignore_notice(n: notice_info): NoticeAction
	{
	return NOTICE_IGNORE;
	}

function file_notice(n: notice_info): NoticeAction
	{
	return NOTICE_FILE;
	}

global notice_tallies: table[string] of count &default = 0;

function tally_notice(s: string)
	{
	++notice_tallies[s];
	}

function tally_notice_type(n: notice_info): NoticeAction
	{
	tally_notice(fmt("%s", n$note));
	return NOTICE_FILE;
	}

event bro_done()
	{
	for ( s in notice_tallies )
		{
		local n = notice_tallies[s];
		alarm fmt("%s (%d time%s)", s, n, n > 1 ? "s" : "");
		}
	}

global notice_file = open_log_file("notice") &redef;

# This handler is useful for processing notices after the notice filters
# have been applied and yielded an NoticeAction.
#
# It's tempting to make the default handler do the logging and
# printing to notice_file, rather than NOTICE.  I hesitate to do that,
# though, because it perhaps could slow down notification, because
# in the absence of event priorities, the event would have to wait
# behind any other already-queued events.
event notice_action(n: notice_info, action: NoticeAction)
	{
	}

# Given a string, returns an escaped version suitable for being
# printed in the colon-separated notice format.  This means that
# (1) any colons are escaped using '\', and (2) any '\'s are
# likewise escaped.
function notice_escape(s: string): string
	{
	s = subst_string(s, "\\", "\\\\");
	return subst_string(s, ":", "\\:");
	}

# Hack to suppress duplicate notice_actions for remote notices.
global suppress_notice_action = F;

function NOTICE(n: notice_info)
	{
	# Fill in some defaults.
	if ( ! n?$id && n?$conn )
		n$id = n$conn$id;

	if ( ! n?$src && n?$id )
		n$src = n$id$orig_h;
	if ( ! n?$dst && n?$id )
		n$dst = n$id$resp_h;

	if ( ! n?$p && n?$id )
		n$p = n$id$resp_p;

	if ( ! n?$src && n?$iconn )
		n$src = n$iconn$orig_h;
	if ( ! n?$dst && n?$iconn )
		n$dst = n$iconn$resp_h;

	if ( ! n?$src_peer )
		n$src_peer = get_event_peer();

	local action = match n using notice_policy;

	if ( n$note in notice_action_filters )
		action = notice_action_filters[n$note](n);

	if ( action != NOTICE_IGNORE && action != NOTICE_FILE )
		{
		local descr = "";
		if ( is_remote_event() )
			{
			if ( n$src_peer$descr != "" )
				descr = fmt("<%s> ", n$src_peer$descr );
			else
				descr = fmt("<%s:%s> ", n$src_peer$host,
							n$src_peer$p);
			}

		alarm fmt("%s %s%s", n$note, descr, n$msg);
		}

	if ( action != NOTICE_IGNORE )
		{
		local t = fmt("%.06f", network_time());
		local notice = fmt("%s", n$note);
		local notice_act = fmt("%s", action);
		local event_src =
			is_remote_event() ?
				(n$src_peer$descr != "" ?
					n$src_peer$descr :
					fmt("%s/%s", n$src_peer$host,
						n$src_peer$p)) : "";
		local src_addr :string;
		local dst_addr :string;
		local src_port :string;
		local dst_port :string;
		if ( n?$invert_conn && n$invert_conn )
			{
			# If we get here it is assumed that a conn object must
			# exist since the boolean flag says to flip the 
			# originator/responder in the conn object.
			dst_addr = (n?$id) ? fmt("%s", n$id$orig_h) : "";
			src_addr = (n?$id) ? fmt("%s", n$id$resp_h) : "";

			dst_port = (n?$id) ? fmt("%s", n$id$orig_p) : "";
			src_port = (n?$id) ? fmt("%s", n$id$resp_p) : "";
			}
		else
			{
			src_addr = n?$src ? fmt("%s", n$src) : "";
			dst_addr = n?$dst ? fmt("%s", n$dst) : "";

			src_port = (n?$id && n$id$orig_h == n$src) ?
						fmt("%s", n$id$orig_p) : "";
			dst_port = (n?$id && n$id$resp_h == n$dst) ?
						fmt("%s", n$id$resp_p) : "";
			}
		local user = n?$user ? notice_escape(n$user) : "";
		local filename = n?$filename ? notice_escape(n$filename) : "";
		local method = n?$method ? notice_escape(n$method) : "";
		local URL = n?$URL ? notice_escape(n$URL) : "";
		local num = n?$n ? fmt("%s", n$n) : "";
		local msg = n?$msg ? notice_escape(n$msg) : "";
		local sub_msg = n?$sub ? notice_escape(n$sub) : "";

		local info =
			fmt("%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s",
				t, notice, notice_act, event_src,
				src_addr, src_port, dst_addr, dst_port,
				user, filename, method, URL, num, msg, sub_msg);

		print notice_file, info;
		}

@ifdef ( IDMEF_support )
	if ( n?$id )
		generate_idmef(n$id$orig_h, n$id$orig_p,
			       n$id$resp_h, n$id$resp_p);
@endif

	event notice_action(n, action);
	}
