package Bro::Report::Alarm;

use strict;
require 5.006_001;
use Bro::Config( '$BRO_CONFIG' );
use Bro::Report qw( trimhostname iptoname swrite time_mdhm time_hms date_md
				standard_deviation getincidentnumber tempfile trimstring );
use Bro::Signature( 'findkeyblocks' );
use Bro::Log::Conn;

use vars qw( $VERSION
			$DEBUG
			$SCANS_MAX_COUNT
			$RPT_CACHE
			$BROHOME
			$SCAN_MAX_BYTES_RCV
			$SCAN_EVENT_REGEX
			@SCAN_EVENT_LIST
			@INCIDENT_EVENT_LIST
			$INCIDENT_EVENT_REGEX
			$REPORTABLE_EVENT_REGEX
			$ALERT_TYPE_SCORES
			$ALERT_TYPE_SCORES_FILE
			$SIGNATURE_ID_SCORES
			$SIGNATURE_ID_SCORES_FILE
			$ALARM_THRESHOLD
			$INCIDENT_TEMP_NUMBER
			$MAX_INCIDENT_CONN_LINES
			$SHOW_UNSUCCESSFUL_INCIDENTS
			$INCIDENT_SHOW_SUB_MESSAGE
			$INCIDENT_SHOW_SIGNATURE );

# $Id: Alarm.pm,v 1.13 2004/10/29 23:42:11 rwinslow Exp $
$VERSION = 1.06;
$SCANS_MAX_COUNT = 20;
$SCAN_MAX_BYTES_RCV = 20480;

$DEBUG = 0;
$INCIDENT_TEMP_NUMBER = 1;

# data for a report will be removed by it's respective function once the data
# has been called to output.  All report data in memory is held in
# variable $RPT_CACHE;

my %REPORT_MAP = ( 'scans' => { input => __PACKAGE__ . '::scans',
						output => __PACKAGE__ . '::output_scans', },
		'incidents' => { input => __PACKAGE__ . '::incident',
					output => __PACKAGE__ . '::output_incident', },
		'scan_summary' => { input => undef,
						output => __PACKAGE__ . '::output_scansummary', },
		'incident_summary' => { input => undef,
						output => __PACKAGE__ . '::output_incidentsummary', },
		'signature_summary' => { input => undef,
						output => __PACKAGE__ . '::output_signaturesummary', },
		'signature_distribution' => { input => __PACKAGE__ . '::signaturedistribution',
						output => __PACKAGE__ . '::output_signaturedistribution', },
		'dropped_packets' => { input => __PACKAGE__ . '::droppedpackets',
						output => __PACKAGE__ . '::output_droppedpackets', },
		);

$ALERT_TYPE_SCORES = {};
$SIGNATURE_ID_SCORES = {};

$ALERT_TYPE_SCORES_FILE = $BRO_CONFIG->{BROSITE} . "/alert_scores";
$SIGNATURE_ID_SCORES_FILE = $BRO_CONFIG->{BROSITE} . "/signature_scores";

# Set the signature score list
setsignaturescores( $SIGNATURE_ID_SCORES_FILE );

# Set the alert_type score list
setalerttypescores( $ALERT_TYPE_SCORES_FILE );

# Current threshold limit, default is 100.
$ALARM_THRESHOLD = 100;

@SCAN_EVENT_LIST = ( 'PortScan', 'PasswordGuessing', 'AddressScan', 
	'MultipleSigResponders', 'MultipleSignatures', );

scaneventlist( @SCAN_EVENT_LIST );

@INCIDENT_EVENT_LIST = ( 'SensitiveSignature', 'FTP_UnexpectedConn',
	'FTP_ExcessiveFilename', 'FTP_PrivPort', 'FTP_BadPort', 'FTP_Sensitive',
	'HTTP_SensitiveURI', 'OutboundTFTP', );

incidenteventlist( @INCIDENT_EVENT_LIST );

$MAX_INCIDENT_CONN_LINES = 25;

# If and how should the unsuccessful incident data be displayed.
# Valid values are 'FIRST INSTANCE', 'ALL'
$SHOW_UNSUCCESSFUL_INCIDENTS = 'FIRST INSTANCE';

# Toggle whether the alarm sub_message should be included in the incident details
$INCIDENT_SHOW_SUB_MESSAGE = 1;

# Toggle wether the actual signature code block should be included in incident details
# when the notice type is SensitiveSignature
$INCIDENT_SHOW_SIGNATURE = 1;

my $DROPPED_PACKETS_REGEX = qr/([[:digit:]]+) packets dropped after filtering, ([[:digit:]]+) received, ([[:digit:]]+) on link/o;

sub scans
{
	my $sub_name = 'scans';
	
	my $_al_struc = $_[0] || return( undef );
	
	my $alert_type = Bro::Log::Alarm::alert_type( $_al_struc );
	if( $alert_type =~ $SCAN_EVENT_REGEX )
	{
		my $src_ip = Bro::Log::Alarm::source_ip( $_al_struc );
		push( @{$RPT_CACHE->{scans}->{$src_ip}->{ALARMS}}, $_al_struc );
	}
	
	return( 1 );
}

sub output_scans
{
	my $sub_name = 'output_scans';
	
	my $format = $_[1];
	my $total_scans = $RPT_CACHE->{scans}->{COUNT};
	my @results;
	my $ret_string;
	my $content = '';
	my $max_reason_length = 63;
	my @heading_names = ( 'Host', 'IP', );
	
	if( ! exists( $RPT_CACHE->{scans} ) )
	{
		return( undef );
	}
	
	if( ! $format )
	{
		$format = <<'END';
  Host:   @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  @<<<<<<<<<<<<<<
  Reason: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  ~       @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  ~       @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

END
	}
		
	# Classify the scans
	classifyscans();
	
	# Reorganize the scan by type
	my %scan_events_by_type;
	foreach my $h_ip( keys( %{$RPT_CACHE->{scans}} ) )
	{
		foreach my $alarm_struc( @{$RPT_CACHE->{scans}->{$h_ip}->{ALARMS}} )
		{
			my $alert_type = Bro::Log::Alarm::alert_type( $alarm_struc );
			push( @{$scan_events_by_type{$alert_type}}, $alarm_struc );
		}
	}
		
	# Make the content
	my $num_scans_output = 0;
	foreach my $event_type( keys( %scan_events_by_type ) )
	{
		foreach my $alarm_struc( @{$scan_events_by_type{$event_type}} )
		{
			my $h_ip = Bro::Log::Alarm::source_ip( $alarm_struc );
			my $h_name = trimhostname( iptoname( $h_ip ), 44, '>' );
			my $message;
			
			# Post process the messages for the following event types
			if( $event_type eq 'MultipleSignatures' )
			{
				$message = Bro::Log::Alarm::count( $alarm_struc ) .
					" different signatures triggered against " .
					Bro::Log::Alarm::destination_ip( $alarm_struc );
			}
			elsif( $event_type eq 'MultipleSigResponders' )
			{
				my $sig_event = Bro::Log::Alarm::message( $alarm_struc );
				my $sigid = Bro::Log::Alarm::sigid( $alarm_struc );
				my $c = Bro::Log::Alarm::count( $alarm_struc );
				$message = "Triggered signature $sigid: $sig_event across $c hosts";
			}
			else
			{
				$message = Bro::Log::Alarm::message( $alarm_struc );
			}
			
			# reduce to a max of three lines at $max_reason_length characters per line
			my @reason = trimstring( $message, $max_reason_length, 3 );
			
			$content .= swrite( $format, $h_name, $h_ip, @reason );
			++$num_scans_output;			
			if( $num_scans_output > $SCANS_MAX_COUNT )
			{
				last;
			}
		}
		if( $num_scans_output > $SCANS_MAX_COUNT )
		{
			last;
		}
	}
	
	if( length( $content ) < 10 )
	{
		$ret_string = "     No data to report\n";
	}
	else
	{
		$ret_string .= $content;
		if( $total_scans > $SCANS_MAX_COUNT )
		{
			my $num_not_displayed = $total_scans - $SCANS_MAX_COUNT;
			$ret_string .= <<"END"
			
    Maximum of $SCANS_MAX_COUNT scans are listed.
    There are another $num_not_displayed that are not displayed.
END
		}
	}
	
	# Clean up some memory
	delete( $RPT_CACHE->{scans} );
	
	return( $ret_string );
}

sub classifyscans
{
	my $sub_name = 'classifyscans';
	
	my $success_scans = 0;
	my $failed_scans = 0;
	
	if( exists( $RPT_CACHE->{scans}->{CLASSIFICATION_TOTALS} ) )
	{
		$success_scans = $RPT_CACHE->{scans}->{CLASSIFICATION_TOTALS}->{'SUCCESSFUL'};
		$failed_scans = $RPT_CACHE->{scans}->{CLASSIFICATION_TOTALS}->{'UNSUCCESSFUL'};
	}
	else
	{
		# Post process some scan types
		foreach my $host( keys( %{$RPT_CACHE->{scans}} ) )
		{
			my @new_alarm_list;
			my %ms;	# MultipleSignatures
			my %msr;	# MultipleSigResponders
			
			foreach my $alarm_struc( @{$RPT_CACHE->{scans}->{$host}->{ALARMS}} )
			{
				my $alert_type = Bro::Log::Alarm::alert_type( $alarm_struc );
				if( $alert_type eq 'MultipleSignatures' )
				{
					# Only keep the highest reported number for each pair
					my $src_ip = Bro::Log::Alarm::source_ip( $alarm_struc );
					my $dst_ip = Bro::Log::Alarm::destination_ip( $alarm_struc );
					if( exists( $ms{"$src_ip$dst_ip"} ) )
					{
						my $cur_count = Bro::Log::Alarm::count( $alarm_struc );
						if( $cur_count > Bro::Log::Alarm::count( $ms{"$src_ip$dst_ip"} ) )
						{
							$ms{"$src_ip$dst_ip"} = $alarm_struc;
						}
					}
					else
					{
						$ms{"$src_ip$dst_ip"} = $alarm_struc;
					}
				}
				elsif( $alert_type eq 'MultipleSigResponders' )
				{
					# Only keep the highest count for each offender
					my $src_ip = Bro::Log::Alarm::source_ip( $alarm_struc );
					if( exists( $msr{"$src_ip"} ) )
					{
						my $cur_count = Bro::Log::Alarm::count( $alarm_struc );
						if( $cur_count > Bro::Log::Alarm::count( $msr{"$src_ip"} ) )
						{
							$msr{"$src_ip"} = $alarm_struc;
						}
					}
					else
					{
						$msr{"$src_ip"} = $alarm_struc;
					}
				}
				else
				{
					push( @new_alarm_list, $alarm_struc );
				}
			}
			
			if( keys( %ms ) > 0 )
			{
				push( @new_alarm_list, values( %ms ) );
			}
			
			if( keys( %msr ) > 0 )
			{
				push( @new_alarm_list, values( %msr ) );
			}
			
			@{$RPT_CACHE->{scans}->{$host}->{ALARMS}} = @new_alarm_list;
		}
		# Figure out if the scan is worth reporting.
		foreach my $h_ip( keys( %{$RPT_CACHE->{scans}} ) )
		{
			my $is_success = 0;
			BLOCK: {
				# Check if there were any connections back to the offender
				if( $RPT_CACHE->{scans}->{$h_ip}->{CONNECTIONS_TO_OFFENDER} )
				{
					if( $DEBUG > 2 )
					{
						warn( "Scan events for $h_ip have be found worthy of reporting.  " . 
							$RPT_CACHE->{scans}->{$h_ip}->{CONNECTIONS_TO_OFFENDER} . 
							" connections were made back to the offender.\n");
					}
					$is_success = 1;
					last BLOCK;
				}
				
				if( exists( $RPT_CACHE->{scans}->{$h_ip}->{BYTES_RCV} ) )
				{
					foreach my $bytes_rcv( keys( %{$RPT_CACHE->{scans}->{$h_ip}->{BYTES_RCV}} ) )
					{
						if( $bytes_rcv > $SCAN_MAX_BYTES_RCV )
						{
							$is_success = 1;
							if( $DEBUG > 2 )
							{
								warn( "Scan events from source $h_ip have been found worthy of reporting." .
									"There was data over $SCAN_MAX_BYTES_RCV sent back to the host\n" );
							}
							last BLOCK;
						}
					}
				}
				
				# Figure out if the bytes sent by the offender is more than $num_deviations
				# deviations out from the standard deviation.
				if( $RPT_CACHE->{scans}->{$h_ip}->{BYTES_SENT} )
				{
					my $num_deviations = 3;
					my $std_dev = standard_deviation( $RPT_CACHE->{scans}->{$h_ip}->{BYTES_SENT} );
					my $max_deviation = sprintf( "%d", $std_dev * $num_deviations );
					my $max_val = 0;

					# Make sure that standard_deviation returned a valid value
					if( defined( $std_dev ) )
					{
						foreach my $num( keys( %{$RPT_CACHE->{scans}->{$h_ip}->{BYTES_SENT}} ) )
						{
							if( $num > $max_deviation )
							{
								$RPT_CACHE->{scans}->{$h_ip}->{SENT_DATA_DEVIATION} = 1;
								if( $DEBUG > 2 )
								{
									$max_val = $num;
								}
								last;
							}

							if( $DEBUG > 2 )
							{
								if( $num > $max_val )
								{
									$max_val = $num;
								}
							}
						}
					}
					else
					{
						# Not worthy of reporting
						$max_deviation = 'undef';
						if( $DEBUG > 2 )
						{
							warn( "Scan events from source $h_ip has been found not worthy of reporting.  " .
								"Not enough data for a standard deviation test.\n" );
						}
					}

					if( $RPT_CACHE->{scans}->{$h_ip}->{SENT_DATA_DEVIATION} )
					{
						# Report !
						$is_success = 1;
						if( $DEBUG > 2 )
						{
							warn( "Scan events from source $h_ip has been found worthy of reporting after " .
								"standard deviation test.  Max deviation was $max_deviation and max value ". 
								"was $max_val\n" );
						}
						last BLOCK;
					}
					else
					{
						# Not worthy of reporting
						delete( $RPT_CACHE->{scans}->{$h_ip} );
						if( $DEBUG > 2 )
						{
							warn( "Scan events from source $h_ip has been found not worthy of reporting after standard deviation test.\n" );
							warn( "Max deviation was $max_deviation and max value was $max_val\n" );
						}
					}
				}
				else
				{
					# Not worthy of reporting
					delete( $RPT_CACHE->{scans}->{$h_ip} );
					if( $DEBUG > 2 )
					{
						warn( "Scan events from source $h_ip has been found not worthy of reporting.\n" );
						warn( "Did not find any data transfered from host and none back to host.\n" );
					}
				}
			} # end BLOCK
			
			if( $is_success )
			{
				++$success_scans;
			}
			else
			{
				++$failed_scans;
			}
		}
		$RPT_CACHE->{scans}->{CLASSIFICATION_TOTALS}->{'SUCCESSFUL'} = $success_scans;
		$RPT_CACHE->{scans}->{CLASSIFICATION_TOTALS}->{'UNSUCCESSFUL'} = $failed_scans;
	}
	
	return( $success_scans, $failed_scans );
}

sub output_scansummary
{
	my $sub_name = 'output_scansummary';
	
	my $ret_string = '';
	
	if( ! exists( $RPT_CACHE->{scans} ) )
	{
		return( undef );
	}
	
	my( $successful_scans, $failed_scans ) = classifyscans();
	my $ret_string = <<"END";
  Scanning Hosts
    Successful            $successful_scans
    Unsuccessful          $failed_scans
END
	
	return( $ret_string );
}

sub scaneventlist
{
	my $sub_name = 'scaneventlist';
	
	my @args = @_;
	
	if( defined( $args[0] ) )
	{
		@SCAN_EVENT_LIST = @args;
		if( @SCAN_EVENT_LIST > 1 )
		{
			my $build_string = join( '|', @args );
			$SCAN_EVENT_REGEX = qr/^(?:$build_string)$/;
		}
		else
		{
			$SCAN_EVENT_REGEX = qr/^$SCAN_EVENT_LIST[0]$/;
		}
	}
	# Rebuild the reportable event regex.
	reportablealerttlist();
	
	return( @SCAN_EVENT_LIST );
}

sub incidenteventlist
{
	my $sub_name = 'incidenteventlist';
	
	my @args = @_;
	
	if( defined( $args[0] ) )
	{
		@INCIDENT_EVENT_LIST = @args;
		if( @INCIDENT_EVENT_LIST > 1 )
		{
			my $build_string = join( '|', @args );
			$INCIDENT_EVENT_REGEX = qr/^(?:$build_string)$/;
		}
		else
		{
			$INCIDENT_EVENT_REGEX = qr/^$INCIDENT_EVENT_LIST[0]$/;
		}
	}
	# Rebuild the reportable event regex.
	reportablealerttlist();
	
	return( @INCIDENT_EVENT_LIST );
}

sub reportablealerttlist
{
	my $sub_name = 'reportablealerttlist';
	my @events = ( @SCAN_EVENT_LIST, @INCIDENT_EVENT_LIST );
	
	if( @events > 1 )
	{
		my $build_string = join( '|', @events );
		$REPORTABLE_EVENT_REGEX = qr/^(?:$build_string)$/;
	}
	else
	{
		$REPORTABLE_EVENT_REGEX = qr/^$events[0]$/;
	}
	
	return( @SCAN_EVENT_LIST, @INCIDENT_EVENT_LIST );
}

sub incident
{
	my $sub_name = 'incident';
	
	my $al_struc = $_[0] || return( undef );
	
	my $alert_type = Bro::Log::Alarm::alert_type( $al_struc ) or return( undef );
	my $src_ip = Bro::Log::Alarm::source_ip( $al_struc ) or return( undef );
	my $dest_ip = Bro::Log::Alarm::destination_ip( $al_struc ) || '';
	
	# Find out if this is an incident worth tracking.
	if( $alert_type =~ $INCIDENT_EVENT_REGEX )
	{
		# Put the incident timestamp in the list of times to watch
		my $timestamp = Bro::Log::Alarm::timestamp( $al_struc );
		push( @{$RPT_CACHE->{'incident'}->{OFFENDERS}->{$src_ip}->{VICTIMS}->{$dest_ip}->{TIMESTAMPS}}, 
			{ TIME => $timestamp, COUNT => 0 } );
		
		my $data_root = $RPT_CACHE->{'incident'}->{OFFENDERS}->{$src_ip}->{VICTIMS}->{$dest_ip};
		
		# Add the alarm to the list
		push( @{$data_root->{ALARMS}}, $al_struc );
		
		if( $alert_type eq 'SensitiveSignature' )
		{
			my $add_score = 0;
			my $sig_id;
			
			# Find the signature id
			if( $sig_id = sigid( $al_struc ) )
			{
				# Get the signature score
				$add_score = signaturescore( $sig_id );
			}
			
			push( @{$data_root->{SCORE}}, $add_score );
			
			# Add the sigid to the hash of known signature alerts
			$RPT_CACHE->{'incident'}->{SIGNATURES}->{$sig_id} = 1;
		}
		else
		{
			# It's some other type of alert.  This is a general approach to
			# alerts with no special handling.
			push( @{$data_root->{SCORE}}, alerttypescore( $alert_type ) );
		}
	}
	else
	{
		return( 0 );
	}
	
	return( 1 );
}

sub output_incident
{
	my $sub_name = 'output_incident';
	
	my $incident_struc = $_[0];
	my $conn_struc_array_ref = $_[1];
	my $_max_output_count = $_[2] || 3000;
	my @results;
	my $ret_string;
	my $incident_header;
	my $connection_pair_format;
	my $alarm_descr_format;
	my $alarm_time_dir_format;
	my $conn_top_format;
	my $conn_format;
	my $detail_legend;
	my $incident_count = 0;
	my $likely_successful = '';
	my $likely_unsuccessful = '';
	my $unknown = '';	
	my %unsuccessful_sig_ids;
	
	if( ! exists( $RPT_CACHE->{incident} ) )
	{
		return( undef );
	}
	
	# Legend to print at the top of the incident detail block
	$detail_legend = <<'END';
                  # legend for connection type #
                  ------------------------------
         C Connection Status
           # number corresponds to alarm triggered by the connection
           * successful connection, otherwise unsuccessful.
         I Initiatator of Connection
           > connection initiated by remote host
           < connection initiated by local host

END
	
	# Start of incident and the unique number associated with it
	$incident_header = <<'END';
-------------------------------------------------------------------------
Incident      @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @>>>>>>>>>>>>>>>>>>>>>
------------------------
END
	
	# connection pair format
	$connection_pair_format = <<'END';
Remote Host:  @<<<<<<<<<<<<<<    @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 Local Host:  @<<<<<<<<<<<<<<    @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

END
	
	# Alarm description output
	$alarm_descr_format = <<'END';
Alarm: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  @<< @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
~     @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
END

	# Alarm time and direction
	$alarm_time_dir_format = <<'END';
      @<<<<<<<<<<<<<                @>>>>>>>>>>>>>> -> @<<<<<<<<<<<<<<
~                                       @>>>>>>>>>> -> @<<<<<<<<<<
END
	
	# Header for the connection details
	$conn_top_format = <<"END";
Connections (only first $MAX_INCIDENT_CONN_LINES after alarm are listed)
-----------
                 time     byte   remote        local   byte
 date   time   duration transfer  port  C   I   port transfer   protocol
----- -------- -------- -------- ------ ------ ----- -------- -----------
END
	
	# Connection detail format
	$conn_format = <<'END';
@<<<< @<<<<<<< @>>>>>>> @>>>>>>> @>>>>> @<<<@> @>>>> @>>>>>>> @>>>>>>>>>>
END
	
	# Check whether signatures are to be included in the report. If so parse
	# and store all signatures which are to be reported on.
	if( $INCIDENT_SHOW_SIGNATURE )
	{
		loadsignaturecode( keys( %{$RPT_CACHE->{incident}->{SIGNATURES}} ) );
	}
	
	foreach my $offender( keys( %{$RPT_CACHE->{incident}->{OFFENDERS}} ) )
	{
		foreach my $__victim( keys( %{$RPT_CACHE->{incident}->{OFFENDERS}->{$offender}->{VICTIMS}} ) )
		{
			my $output;
			my $__data = $RPT_CACHE->{incident}->{OFFENDERS}->{$offender}->{VICTIMS}->{$__victim};
			if( $__data->{CLASS} eq 'LIKELY SUCCESSFUL' )
			{
				$output = \$likely_successful;
			}
			elsif( $__data->{CLASS} eq 'UNKNOWN' )
			{
				$output = \$unknown;
			}
			elsif( $__data->{CLASS} eq 'LIKELY UNSUCCESSFUL' and 
				$SHOW_UNSUCCESSFUL_INCIDENTS )
			{
				$output = \$likely_unsuccessful;
				
				if( $SHOW_UNSUCCESSFUL_INCIDENTS eq 'FIRST INSTANCE' )
				{
					my @new_alarm_list;
					foreach my $alarm( @{$__data->{ALARMS}} )
					{
						if( Bro::Log::Alarm::alert_type( $alarm ) eq 'SensitiveSignature' )
						{
							if( ! exists( $unsuccessful_sig_ids{Bro::Log::Alarm::sigid( $alarm )} ) )
							{
								$unsuccessful_sig_ids{Bro::Log::Alarm::sigid( $alarm )} = 1;
								push( @new_alarm_list, $alarm );
							}
						}
						else
						{
							push( @new_alarm_list, $alarm );
						}
					}
					
					if( scalar( @new_alarm_list ) < 1 )
					{
						next;
					}
					else
					{
						$__data->{ALARMS} = \@new_alarm_list;
					}
				}
			}
			else
			{
				next;
			}
			
			my @alarms;
			my %alarm_times;
			# Sort the alarms in ascending order (probably already sorted but make sure)
			for( my $a_idx = 0; $a_idx < @{$__data->{ALARMS}}; ++$a_idx )
			{
				my $timestamp = Bro::Log::Alarm::timestamp( $__data->{ALARMS}->[$a_idx] );
				push( @{$alarm_times{$timestamp}}, $a_idx );
			}
			
			foreach my $ts( sort( {$a <=> $b} keys( %alarm_times ) ) )
			{
				foreach my $idx( @{$alarm_times{$ts}} )
				{
					push( @alarms, $__data->{ALARMS}->[$idx] );
				}
			}
			
			# Store conn strucs for writting later
			my @conn_strucs;
			
			for( my $i_idx = 0; $i_idx < @alarms; ++$i_idx )
			{	
				my $begin_timestamp = Bro::Log::Alarm::timestamp( $alarms[$i_idx] );
				my $end_timestamp;
				if( defined( $alarms[$i_idx + 1] ) )
				{
					$end_timestamp = $alarms[$i_idx + 1];
				}
				push( @conn_strucs, incidentconndata( $offender, $__victim, $begin_timestamp, $end_timestamp ) );
			}
			
			my $incident_id = $BRO_CONFIG->{BRO_SITE_NAME} . '-' . sprintf( "%06d", getincidentnumber() );
			my $victim_hostname = iptoname( $__victim );
			my $offender_hostname = iptoname( $offender );
			my %alarm_times;
			my $last_alarm_idx = 0;
			my $conn_details_output;
			my %conn_reference;
			my $conn_ref_count = 1;
			
			$$output .= swrite( $incident_header, $incident_id, $__data->{CLASS} );
			
			# If there is no connection data then skip the conncetion data ties
			if( ! $conn_strucs[0] )
			{
				# No connection data
				
				$conn_details_output .= $conn_top_format;
				$conn_details_output .= "    No connection data available\n";
			}
			else
			{

				my $local_ip;
				my $local_hostname;
				my $remote_ip;
				my $remote_hostname;
				
				# Figure out which host is local, victim or offender
				my $source_net = Bro::Log::Conn::source_network( $conn_strucs[0] );
				if( $source_net eq 'L' )
				{
					if( Bro::Log::Conn::source_ip( $conn_strucs[0] ) eq $offender )
					{
						$remote_ip = $__victim;
						$remote_hostname = $victim_hostname;
						$local_ip = $offender;
						$local_hostname = $offender_hostname;
					}
					else
					{
						$remote_ip = $offender;
						$remote_hostname = $offender_hostname;
						$local_ip = $__victim;
						$local_hostname = $victim_hostname;
					}
					
					# Trim the hostnames down if needed
					$local_hostname = trimhostname( $local_hostname ,40 , '<' );
					$remote_hostname = trimhostname( $remote_hostname ,40 , '>' );
				}
				else
				{
					if( Bro::Log::Conn::source_ip( $conn_strucs[0] ) eq $offender )
					{
						$remote_ip = $offender;
						$remote_hostname = $offender_hostname;
						$local_ip = $__victim;
						$local_hostname = $victim_hostname;
					}
					else
					{
						$remote_ip = $__victim;
						$remote_hostname = $victim_hostname;
						$local_ip = $offender;
						$local_hostname = $offender_hostname;
					}
				}

				$$output .= swrite( $connection_pair_format, $remote_ip, $remote_hostname,
								$local_ip, $local_hostname );

				# Delay writing of the details to the returned string
				$conn_details_output .= $conn_top_format;

				foreach my $conn_struc( @conn_strucs )
				{
					my $timestamp = Bro::Log::Conn::timestamp( $conn_struc );
					my $date = date_md( $timestamp );
					my $time = time_hms( $timestamp );
					my $duration = Bro::Log::Conn::duration( $conn_struc, 'raw' );
					my $service = Bro::Log::Conn::service( $conn_struc );
					my $remote_bytes;
					my $remote_port;
					my $local_bytes;
					my $local_port;
					my $direction;

					if( Bro::Log::Conn::source_ip( $conn_struc ) eq $remote_ip )
					{
						$remote_bytes = Bro::Log::Conn::source_bytes( $conn_struc, 'raw' );
						$remote_port = Bro::Log::Conn::source_port( $conn_struc );
						$local_bytes = Bro::Log::Conn::destination_bytes( $conn_struc, 'raw' );
						$local_port = Bro::Log::Conn::destination_port( $conn_struc );
						$direction = ' >';
					}
					else
					{
						$remote_bytes = Bro::Log::Conn::destination_bytes( $conn_struc, 'raw' );
						$remote_port = Bro::Log::Conn::destination_port( $conn_struc );
						$local_bytes = Bro::Log::Conn::source_bytes( $conn_struc, 'raw' );
						$local_port = Bro::Log::Conn::source_port( $conn_struc );
						$direction = '< ';
					}

					my $conn_stat;

					# Check if any of the alarms have references in the other field of
					# the connection data and the timestamp of the alarm falls within
					# the connection period.
					for( my $i = 0; $i < @alarms; ++$i )
					{
						my $alarm = $alarms[$i];
						my $alarm_time = Bro::Log::Alarm::timestamp( $alarm );
						if( Bro::Log::Conn::range( $conn_struc, $alarm_time, '.03' ) )
						{
							if( Bro::Log::Alarm::alert_type( $alarm ) eq 'SensitiveSignature' )
							{
								my $other_field = Bro::Log::Conn::other( $conn_struc );
								my $sigid = sigid( $alarm );
								if( $other_field =~ m/$sigid/ )
								{
									$conn_reference{$i} = $conn_ref_count;
									$conn_stat = $conn_ref_count;
								}
							}
						}
					}

					if( $conn_stat )
					{
						$conn_stat = $conn_ref_count;
						++$conn_ref_count;
					}
					elsif( Bro::Log::Conn::connectsucceed( $conn_struc ) )
					{
						$conn_stat = '*';
					}
					else
					{
						$conn_stat = '';
					}

					$conn_details_output .= swrite( $conn_format, $date, $time, $duration,
						$remote_bytes, $remote_port, $conn_stat, $direction,
						$local_port, $local_bytes, $service );
				}
			}
			
			# Now go back through and create the alarms section
			for( my $a_idx = 0; $a_idx < @alarms; ++$a_idx )
			{
				my $reference_idx;
				my $offender_port;
				my $__victim_port;
				
				# Get the source/dest ports and figure out which is local/remote
				if( Bro::Log::Alarm::source_ip( $alarms[$a_idx] ) eq $offender )
				{
					$offender_port = Bro::Log::Alarm::source_port( $alarms[$a_idx] );
					$__victim_port = Bro::Log::Alarm::destination_port( $alarms[$a_idx] );
				}
				else
				{
					$__victim_port = Bro::Log::Alarm::source_port( $alarms[$a_idx] );
					$offender_port = Bro::Log::Alarm::destination_port( $alarms[$a_idx] );
				}
				
				if( exists( $conn_reference{$a_idx} ) )
				{
					$reference_idx = $conn_reference{$a_idx};
				}
				else
				{
					$reference_idx = '';
				}
				
				my $alert_type = Bro::Log::Alarm::alert_type( $alarms[$a_idx] );
				my $event_msg;
				if( $alert_type eq 'SensitiveSignature' )
				{
					$event_msg .= Bro::Log::Alarm::sigid( $alarms[$a_idx] );
					$event_msg .= ": " . Bro::Log::Alarm::message( $alarms[$a_idx] );
				}
				else
				{
					$event_msg .= Bro::Log::Alarm::message( $alarms[$a_idx] );
				}
				
				# Make sure that the event message has some value otherwise
				# add something default
				if( ! $event_msg )
				{
					$event_msg = '(no event message available)';
				}
				
				my( $message1, $message2 ) = trimstring( $event_msg, 67, 2 );
				my $time = date_md( Bro::Log::Alarm::timestamp( $alarms[$a_idx] ) ) .
					' '. time_hms( Bro::Log::Alarm::timestamp( $alarms[$a_idx] ) );
				$$output .= swrite( $alarm_descr_format, $alert_type, $reference_idx,
								$message1, $message2 );
				$$output .= swrite( $alarm_time_dir_format, $time, $offender,
					$__victim, $offender_port, $__victim_port );
								
				# Attach the signature code if the notice type is SensitiveSignature
				# and $INCIDENT_SHOW_SIGNATURE is true
				if( $alert_type eq 'SensitiveSignature' and $INCIDENT_SHOW_SIGNATURE )
				{
					$$output .= "      signature code:\n";
					my @parts = split( /\n/, signaturecode( Bro::Log::Alarm::sigid( $alarms[$a_idx] ) ) );
					foreach my $_line( @parts )
					{
						$$output .= "        $_line\n";
					}
				}
				
				# Attach sub message if available and option set
				if( Bro::Log::Alarm::sub_message( $alarms[$a_idx] ) and $INCIDENT_SHOW_SUB_MESSAGE )
				{
					if( $alert_type eq 'SensitiveSignature' )
					{
						$$output .= "      payload:\n";	
					}
					else
					{
						$$output .= "      sub message:\n";	
					}
					
					foreach my $sub_message( trimstring( Bro::Log::Alarm::sub_message( $alarms[$a_idx] ), 66 ) )
					{
						$$output .= "        $sub_message\n";
					}
				}
				# Add a newline between alarms
				$$output .= "\n";
			}
			
			$$output .= $conn_details_output;
			
			$$output .= "-----------------------------\n\n\n";
			
			++$incident_count;
		}
	}
	
	if( $incident_count < 1 )
	{
		$ret_string = "     No data to report\n";
	}
	else
	{
		$ret_string .= $detail_legend . $likely_successful . $unknown . $likely_unsuccessful;
	}
	
	# Clean up some memory
	delete( $RPT_CACHE->{incident} );
	
	return( $ret_string );
}

sub classifyincidents
{
	my $sub_name = 'classifyincidents';
	
	my $success = 0;
	my $unsuccess = 0;
	my $unknown = 0;
	
	if( ! exists( $RPT_CACHE->{incident}->{OFFENDERS} ) )
	{
		return( undef );
	}
	
	if( exists( $RPT_CACHE->{incident}->{CLASSIFICATION_TOTALS} ) )
	{
		$success = $RPT_CACHE->{incident}->{CLASSIFICATION_TOTALS}->{'LIKELY SUCCESSFUL'};
		$unsuccess = $RPT_CACHE->{incident}->{CLASSIFICATION_TOTALS}->{'LIKELY UNSUCCESSFUL'};
		$unknown = $RPT_CACHE->{incident}->{CLASSIFICATION_TOTALS}->{'UNKNOWN'};
	}
	else
	{
		foreach my $offender( keys( %{$RPT_CACHE->{incident}->{OFFENDERS}} ) )
		{
			if( ! exists( $RPT_CACHE->{incident}->{OFFENDERS}->{$offender}->{VICTIMS} ) )
			{
				if( $DEBUG > 2 )
				{
					warn( __PACKAGE__ . "::$sub_name, No victims listed for offender $offender\n" );
				}
				next;
			}
			
			while( my ( $__victim, $__data ) = each( %{$RPT_CACHE->{incident}->{OFFENDERS}->{$offender}->{VICTIMS}} ) )
			{
				my $BIGGEST_SCORE;	# store the largest score found for incident
				
				if( exists( $__data->{SCORE} ) )
				{
					$BIGGEST_SCORE = ( sort( {$b <=> $a} @{$__data->{SCORE}} ) )[0];
				}
				else
				{
					if( $DEBUG > 2 )
					{
						warn( __PACKAGE__ . "::$sub_name, No scores set for offender $offender.  Defaulting to 0\n" );
					}
					
					push( @{$__data->{SCORE}}, 0 );
					$BIGGEST_SCORE = 0;
				}
				
				my $likelyhood;
				if( $BIGGEST_SCORE >= $ALARM_THRESHOLD )
				{
					$__data->{CLASS} = 'LIKELY SUCCESSFUL';
					++$success;
				}
				elsif( $BIGGEST_SCORE == 0 )
				{
					$__data->{CLASS} = 'UNKNOWN';
					++$unknown;
				}
				else
				{
					$__data->{CLASS} = 'LIKELY UNSUCCESSFUL';
					++$unsuccess;
				}
			}
		}
		$RPT_CACHE->{incident}->{CLASSIFICATION_TOTALS}->{'LIKELY SUCCESSFUL'} = $success;
		$RPT_CACHE->{incident}->{CLASSIFICATION_TOTALS}->{'LIKELY UNSUCCESSFUL'} = $unsuccess;
		$RPT_CACHE->{incident}->{CLASSIFICATION_TOTALS}->{'UNKNOWN'} = $unknown;
	}	
	
	return( $success, $unsuccess, $unknown );
}

sub output_incidentsummary
{
	my $sub_name = 'output_incidentsummary';
	
	# NOTE: An incident is defined as one or more alarms that occur between two ip
	# addresses.
	
	my $ret_string = '';
	
	if( ! exists( $RPT_CACHE->{incident} ) )
	{
		return( undef );
	}
	
	my( $success, $unsuccess, $unknown ) = classifyincidents();
	
	$ret_string = <<"END";
  Incidents
    Likely Successful     $success
    Unknown               $unknown
    Likely Unsuccessful   $unsuccess
END
	
	return( $ret_string );
}

sub output_signaturesummary
{
	my $sub_name = 'output_signaturesummary';
	
	my $ret_string = '';
	my $total_sigs = 0;
	my $unique_sigs = 0;
	my $unique_sources = 0;
	my $unique_destinations = 0;
	my $src_dest_pairs = 0;
	
	my %source_dest_pairs;
	my %sources;
	my %dests;
	
	if( ! $RPT_CACHE->{signaturedistribution} )
	{
		return( undef );
	}
	
	foreach my $sigid( keys( %{$RPT_CACHE->{signaturedistribution}} ) )
	{
		$total_sigs += $RPT_CACHE->{signaturedistribution}->{$sigid}->{COUNT};
		++$unique_sigs;
		foreach my $src( keys( %{$RPT_CACHE->{signaturedistribution}->{$sigid}->{SOURCE}} ) )
		{
			$sources{$src} = 1;
		}
		
		foreach my $dest( keys( %{$RPT_CACHE->{signaturedistribution}->{$sigid}->{DEST}} ) )
		{
			$dests{$dest} = 1;
		}
		
		foreach my $pair( keys( %{$RPT_CACHE->{signaturedistribution}->{$sigid}->{PAIR}} ) )
		{
			$source_dest_pairs{$pair} = 1;
		}
	}
	
	if( my $total = keys( %sources ) )
	{
		$unique_sources = $total;
	}
	
	if( my $total = keys( %dests ) )
	{
		$unique_destinations = $total;
	}
	
	if( my $total = keys( %source_dest_pairs ) )
	{
		$src_dest_pairs = $total;
	}
	
	my $ret_string = <<"END";
  Signature Summary
    Total signatures          $total_sigs
    Unique signatures         $unique_sigs
    Unique sources            $unique_sources
    Unique destinations       $unique_destinations
    Unique source/dest pairs  $src_dest_pairs
END
		
	return( $ret_string );
}

sub signaturedistribution
{
	my $sub_name = 'signaturedistribution';
	
	my $alarm_struc = $_[0] || return( undef );
	
	if( Bro::Log::Alarm::alert_type( $alarm_struc ) eq 'SensitiveSignature' )
	{
		my $sigid = Bro::Log::Alarm::sigid( $alarm_struc );
		my $src_ip = Bro::Log::Alarm::source_ip( $alarm_struc );
		my $dst_ip = Bro::Log::Alarm::destination_ip( $alarm_struc );
		
		if( $sigid and $src_ip and $dst_ip )
		{
			++$RPT_CACHE->{signaturedistribution}->{$sigid}->{SOURCE}->{$src_ip};
			++$RPT_CACHE->{signaturedistribution}->{$sigid}->{DEST}->{$dst_ip};
			++$RPT_CACHE->{signaturedistribution}->{$sigid}->{PAIR}->{"$src_ip-$dst_ip"};
			++$RPT_CACHE->{signaturedistribution}->{$sigid}->{COUNT};
		}
	}
}

sub output_signaturedistribution
{
	my $sub_name = 'output_signaturedistribution';
	
	my $max_output = $_[0] || 20;
	my %reversed_hash;
	my $ret_string = '';
	my @ordered_list;
	
	my $signaturecount_header = <<'END';
                                        Unique      Unique
  Signature ID                Count     Sources     Dests    Unique Pairs
  ------------------------  ---------  ---------  ---------  ------------
END
	my $signaturecount_format = <<'END';
  @<<<<<<<<<<<<<<<<<<<<<<<  @<<<<<<<<  @<<<<<<<<  @<<<<<<<<  @<<<<<<<<<<<
END

	if( ! exists( $RPT_CACHE->{signaturedistribution} ) )
	{
		return( undef );
	}
	
	# Reverse the hash
	foreach my $sigid( keys( %{$RPT_CACHE->{signaturedistribution}} ) )
	{
		push( @{$reversed_hash{$RPT_CACHE->{signaturedistribution}->{$sigid}->{COUNT}}}, $sigid );
	}
	
	# Sort and then set to $ret_string
	@ordered_list = sort( { $b<=>$a } keys( %reversed_hash ) );
	
	my $i = 0;
	while( defined( my $count = shift( @ordered_list ) ) and $i < $max_output )
	{
		foreach my $sigid( @{$reversed_hash{$count}} )
		{
			my $source_count = 0;
			my $dest_count = 0;
			my $pair_count = 0;
			
			$source_count = keys( %{$RPT_CACHE->{signaturedistribution}->{$sigid}->{SOURCE}} );
			$dest_count = keys( %{$RPT_CACHE->{signaturedistribution}->{$sigid}->{DEST}} );
			$pair_count = keys( %{$RPT_CACHE->{signaturedistribution}->{$sigid}->{PAIR}} );
			
			$ret_string .= swrite( $signaturecount_format,
							$sigid,
							$count,
							$source_count,
							$dest_count,
							$pair_count, );
			++$i;
			if( !( $i < $max_output ) )
			{
				last;
			}
		}
	}
	
	if( length( $ret_string ) < 1 )
	{
		$ret_string = "  No data to report\n";
	}
	else
	{
		$ret_string = $signaturecount_header . $ret_string;
	}
	
	return( $ret_string );
}

sub droppedpackets
{
	my $sub_name = 'droppedpackets';
	
	my $alarm_struc = $_[0] || return( undef );
	
	if( Bro::Log::Alarm::alert_type( $alarm_struc ) eq 'DroppedPackets' )
	{
		if( Bro::Log::Alarm::message( $alarm_struc ) =~ $DROPPED_PACKETS_REGEX )
		{
			my $dropped = $1 || 0;
			my $packets_since_last_notice = $3 || 0;
			$RPT_CACHE->{droppedpackets}->{DROPPED} += $dropped;
			$RPT_CACHE->{droppedpackets}->{RECIEVED} += $packets_since_last_notice;
		}
		else
		{
			return( undef );
		}
		return( 1 );
	}
	else
	{
		return( 0 );
	}
}

sub output_droppedpackets
{
	my $sub_name = 'output_droppedpackets';
	
	my $dropped;
	my $recieved;
	my $percent_dropped;
	my $ret_string;
	
	if( exists( $RPT_CACHE->{droppedpackets} ) and 
		$RPT_CACHE->{droppedpackets}->{RECIEVED} > 0 )
	{
		$dropped = $RPT_CACHE->{droppedpackets}->{DROPPED};
		$recieved = $RPT_CACHE->{droppedpackets}->{RECIEVED} + $dropped;
		
		if( $dropped > 0 )
		{
			$percent_dropped = sprintf( "%.4f", ( $dropped / $recieved ) * 100 );
		}
		else
		{
			$percent_dropped = '0%';
		}
	}
	else
	{
		return( undef );
	}
	
	$ret_string = <<"END";
  Dropped Packets
    Packets Recieved: $recieved
    Packets Dropped:  $dropped
    Percent Dropped: $percent_dropped\%
END
	
	# Clean up some memory
	delete( $RPT_CACHE->{droppedpackets} );
	
	return( $ret_string );

}

sub sigid
{
	my $sub_name = 'sigid';
	
	my $_alarm_struc = $_[0];
	my $ret_val;
	
	my $ret_val = Bro::Log::Alarm::filename( $_alarm_struc );
	
	return( $ret_val );	
}

sub setsignaturescores
{
	my $sub_name = 'setsignaturescores';
	
	my $_sigid = $_[0];
	my $ret_hash = {};
	
	if( open( INFILE, $SIGNATURE_ID_SCORES_FILE ) )
	{
		while( defined( my $line = <INFILE> ) )
		{
			if( $line !~ m/^[[:space:]]*\#/ )
			{
				my( $key, $val, $junk ) = split( " ", $line, 3 );
				$ret_hash->{$key} = $val;
			}
		}
	}
	else
	{
		warn( "Unable to open signature score file at $SIGNATURE_ID_SCORES_FILE\n" );
		return( undef );
	}
	
	# Capitalize the _DEFAULT_ parameter if not already
	if( exists( $ret_hash->{'_default_'} ) )
	{
		$ret_hash->{'_DEFAULT_'} = $ret_hash->{'_default_'};
		delete( $ret_hash->{'_default_'} );
	}
	
	# Make sure that a default has been set
	if( ! exists( $ret_hash->{'_DEFAULT_'} ) or
		$ret_hash->{'_DEFAULT_'} !~ m/^[[:digit:]]+$/ )
	{
		$ret_hash->{'_DEFAULT_'} = 5;
	}
	
	close( INFILE );
	
	# Set the package variable to the signature scores taken from the file
	$SIGNATURE_ID_SCORES = $ret_hash;
	
	if( $DEBUG > 3 )
	{
		warn( "Current scores for signature ids\n" );
		while( my( $key, $val ) = each( %{$ret_hash} ) )
		{
			warn( "SIGID: $key => SCORE: $val\n" );
		}
		warn( "\n" );
	}
	
	return( $ret_hash );
}

sub setalerttypescores
{
	my $sub_name = 'setalerttypescores';
	
	my $_sigid = $_[0];
	my $ret_hash = {};
	
	if( open( INFILE, $ALERT_TYPE_SCORES_FILE ) )
	{
		while( defined( my $line = <INFILE> ) )
		{
			if( $line !~ m/^[[:space:]]*\#/ )
			{
				my( $key, $val, $junk ) = split( " ", $line, 3 );
				$ret_hash->{$key} = $val;
			}
		}
	}
	else
	{
		warn( "Unable to open alert_type score file at $ALERT_TYPE_SCORES_FILE\n" );
		return( undef );
	}
	
	# Capitalize the _DEFAULT_ parameter if not already
	if( exists( $ret_hash->{'_default_'} ) )
	{
		$ret_hash->{'_DEFAULT_'} = $ret_hash->{'_default_'};
		delete( $ret_hash->{'_default_'} );
	}
	
	# Make sure that a default has been set
	if( ! exists( $ret_hash->{'_DEFAULT_'} ) or
		$ret_hash->{'_DEFAULT_'} !~ m/^[[:digit:]]+$/ )
	{
		$ret_hash->{'_DEFAULT_'} = 0;
	}
	
	close( INFILE );
	
	# Set the package variable to the signature scores taken from the file
	$ALERT_TYPE_SCORES = $ret_hash;
	
	if( $DEBUG > 3 )
	{
		warn( "Current scores for alert types\n" );
		while( my( $key, $val ) = each( %{$ret_hash} ) )
		{
			warn( "ALERT_TYPE: $key => SCORE: $val\n" );
		}
		warn( "\n" );
	}
	
	return( $ret_hash );
}

sub signaturescore
{
	my $sub_name = 'signaturescore';
	
	my $sigid = $_[0] || '_DEFAULT_';
	my $ret_val;
	
	if( exists( $SIGNATURE_ID_SCORES->{$sigid} ) )
	{
		$ret_val = $SIGNATURE_ID_SCORES->{$sigid};
	}
	else
	{
		$ret_val = $SIGNATURE_ID_SCORES->{'_DEFAULT_'};
	}
	
	return( $ret_val );
}


sub alerttypescore
{
	my $sub_name = 'alerttypescore';
	
	my $alert_type = $_[0] || '_DEFAULT_';
	my $ret_val;
	
	if( exists( $ALERT_TYPE_SCORES->{$alert_type} ) )
	{
		$ret_val = $ALERT_TYPE_SCORES->{$alert_type};
	}
	else
	{
		$ret_val = $ALERT_TYPE_SCORES->{'_DEFAULT_'};
	}
	
	return( $ret_val );
}
sub addconndata
{
	my $sub_name = 'addconndata';
	# The purpose of this function is to add additional data to report
	# parts that can use it.  The data can be used to weight probabilities
	# or add data to a report.
	
	# Filehandle which contains connection data.
	my $conn_struc = $_[0] || return( undef );	# ref to Bro::Log::Conn struc
	my $conn_line = $_[1] || return( undef );	# ref to string
	
	# Scans really only mean something if a successful connection was made
	# or an annomylous conection out of the scan was made.
	my $src_ip = Bro::Log::Conn::source_ip( $conn_struc );
	my $dest_ip = Bro::Log::Conn::destination_ip( $conn_struc );
	my $service = Bro::Log::Conn::service( $conn_struc );
	my $success_connect = Bro::Log::Conn::connectsucceed( $conn_struc );
	my $src_net = Bro::Log::Conn::source_network( $conn_struc );
	if( exists( $RPT_CACHE->{scans}->{$src_ip} ) and
		$RPT_CACHE->{scans}->{$src_ip}->{CONNECTIONS_TO_OFFENDER} )
	{
		if( $service !~ m/^dns|ident$/ and
				$src_net eq 'L' and
				$success_connect )
		{
			++$RPT_CACHE->{scans}->{$dest_ip}->{CONNECTIONS_TO_OFFENDER};
		}
	}
	# If a scanner and not a service to ignore and bytes >= 0 then store
	# the byte count for a post collection standard deviation test.
	elsif( exists( $RPT_CACHE->{scans}->{$src_ip} ) )
	{
		my $src_bytes = Bro::Log::Conn::source_bytes( $conn_struc );
		my $dest_bytes = Bro::Log::Conn::destination_bytes( $conn_struc );
		if( $service !~ m/^dns|ident$/ )
		{
			if( $src_bytes >= 0 )
			{
				++$RPT_CACHE->{scans}->{$src_ip}->{BYTES_SENT}->{$src_bytes};
				++$RPT_CACHE->{scans}->{$src_ip}->{BYTES_RCV}->{$dest_bytes};
			}
		}
	}
	# If a scanner is the destination and our local_net is the orginator
	# of the connection this may be very interesting.
	elsif( exists( $RPT_CACHE->{scans}->{$dest_ip} ) and
		$service !~ m/^dns|ident$/ and
		$src_net eq 'L' )
	{
		++$RPT_CACHE->{scans}->{$dest_ip}->{CONNECTIONS_TO_OFFENDER};
	}
	
	# Add connection data to incidents
	my $offender_address;
	my $victim_address;
	if( exists( $RPT_CACHE->{incident}->{OFFENDERS}->{$src_ip} ) )
	{
		# Does the connection data contain the victim address
		if( exists( $RPT_CACHE->{incident}->{OFFENDERS}->{$src_ip}->{VICTIMS}->{$dest_ip} ) )
		{
			$offender_address = $src_ip;
			$victim_address = $dest_ip;
		}
	}
	elsif( exists( $RPT_CACHE->{incident}->{OFFENDERS}->{$dest_ip} ) )
	{
		# Does the connection data contain the victim address
		if( exists( $RPT_CACHE->{incident}->{OFFENDERS}->{$dest_ip}->{VICTIMS}->{$src_ip} ) )
		{
			$offender_address = $dest_ip;
			$victim_address = $src_ip;
			if( $service !~ m/^dns|ident$/ )
			{
				# This means there was a connection from the victim back to the 
				# offender.  Most likely very interesting.
				push( @{$RPT_CACHE->{incident}->{OFFENDERS}->{$dest_ip}->{VICTIMS}->{$src_ip}->{SCORE}}, 100 );
			}
		}
	}
	
	if( $offender_address )
	{
		if( ! $victim_address )
		{
			$victim_address = '';
		}
		
		my $temp_fh = $RPT_CACHE->{incident}->{'TEMP_FILE_HANDLE'};
		my $incident_data = $RPT_CACHE->{incident}->{OFFENDERS}->{$offender_address}->{VICTIMS}->{$victim_address};
		my $cur_timestamp_idx = $incident_data->{TIMESTAMP_INDEX};
		my $timestamp = Bro::Log::Conn::timestamp( $conn_struc );
		my $duration = Bro::Log::Conn::duration( $conn_struc );
		my $end_of_conn = 9999999999;
		# Connection data varies wildly.  If no duration is given then assume 10 seconds
		if( $duration > -1 )
		{
			$end_of_conn = $timestamp + $duration;
		}
		else
		{
			$end_of_conn = $timestamp + 10;
		}
		if( $incident_data->{TIMESTAMPS}->[$cur_timestamp_idx]->{COUNT} < $MAX_INCIDENT_CONN_LINES + 100 and
			( Bro::Log::Conn::range( $conn_struc, $incident_data->{TIMESTAMPS}->[$cur_timestamp_idx]->{TIME}, .03 ) or
			 $timestamp > $incident_data->{TIMESTAMPS}->[$cur_timestamp_idx]->{TIME} ) )
		{
			if( exists( $incident_data->{TIMESTAMPS}->[$cur_timestamp_idx + 1] ) )
			{
				if( $timestamp >= $incident_data->{TIMESTAMPS}->[$cur_timestamp_idx + 1]->{TIME} )
				{
					$incident_data->{TIMESTAMP_INDEX} = $cur_timestamp_idx + 1;
				}
			}
			
			if( ! defined( $incident_data->{TIMESTAMPS}->[$cur_timestamp_idx]->{OLDEST} ) or
				$timestamp > $incident_data->{TIMESTAMPS}->[$cur_timestamp_idx]->{OLDEST} )
			{
				$incident_data->{TIMESTAMPS}->[$cur_timestamp_idx]->{OLDEST} = $timestamp;
			}
			print $temp_fh $$conn_line;
		}
		# See if the conn data is out of order and possibly matches a previous
		# incident timestamp
		else
		{
			foreach my $idx( @{$incident_data->{TIMESTAMPS}} )
			{
				if( ( Bro::Log::Conn::range( $conn_struc, $idx->{TIME}, .03 ) or
					$timestamp >= $idx->{TIME} ) and
					$timestamp < $idx->{OLDEST} )
				{
					print $temp_fh $$conn_line;
				}
			}
		}
	}
	
	return( 1 );
}

sub reportableoffense
{
	my $sub_name = 'reportableoffense';
	
	my $alarm_struc = $_[0] || return( undef );
	
	if( Bro::Log::Alarm::alert_type( $alarm_struc ) =~ $REPORTABLE_EVENT_REGEX )
	{
		return( 1 );
	}
	else
	{
		return( 0 );
	}
}

sub tempincidentfile
{
	my $sub_name = 'tempincidentfile';
	
	my $arg = $_[0];
	
	if( $arg eq 'close' )
	{
		close( $RPT_CACHE->{incident}->{TEMP_FILE_HANDLE} );
		delete( $RPT_CACHE->{incident}->{TEMP_FILE_HANDLE} );
		return( 1 );
	}
	
	if( ! exists( $RPT_CACHE->{incident}->{TEMP_FILE_NAME} ) )
	{
		$RPT_CACHE->{incident}->{TEMP_FILE_NAME} = 
			tempfile( 'add', "incident_conn_data." );
		
		if( $DEBUG > 2 )
		{
			warn( "Created incident temp file " . $RPT_CACHE->{incident}->{TEMP_FILE_NAME} . "\n" );
		}
	}
	
	if( exists( $RPT_CACHE->{incident}->{TEMP_FILE_HANDLE} ) )
	{
		return( $RPT_CACHE->{incident}->{TEMP_FILE_HANDLE} );
	}
	else
	{
		if( open( INCTEMPOUT, ">>" .$RPT_CACHE->{incident}->{TEMP_FILE_NAME} ) )
		{
			$RPT_CACHE->{incident}->{TEMP_FILE_HANDLE} = \*INCTEMPOUT;
			return( $RPT_CACHE->{incident}->{TEMP_FILE_HANDLE} );
		}
		else
		{
			warn( __PACKAGE__ . "::$sub_name, Unable to open temp file " . $RPT_CACHE->{incident}->{TEMP_FILE_NAME} . " for writting\n" );
			return( undef );
		}
	}
}

sub incidentconndata
{
	my $sub_name = 'incidentconndata';
	
	my $offender = $_[0] || return( undef );
	my $victim = $_[1] || return( undef );
	my $start_time = $_[2] || -1;
	my $end_time = $_[3] || 9999999999;
	
	my @matching_lines;
	my %times;
	my $idx = 0;
	my @ret_strucs;
	
	if( ! $RPT_CACHE->{incident}->{TEMP_FILE_NAME} )
	{
		warn( "No incident temp file exists.  Unable to add connection data to incidents.\n" );
		return( undef );
	}
	
	if( open( INFILE, $RPT_CACHE->{incident}->{TEMP_FILE_NAME} ) )
	{
		while( defined( my $line = <INFILE> ) and $idx < $MAX_INCIDENT_CONN_LINES )
		{
			if( $line =~ m/$victim/ )
			{
				if( $line =~ m/$offender/ )
				{
					my $conn_struc = Bro::Log::Conn::new( \$line );
					my( $conn_start, $conn_end ) = Bro::Log::Conn::range( $conn_struc );
					
					if( ! $conn_end )
					{
						$conn_end = $conn_start + 10;
					}
					
					if( ( $conn_start <= $start_time and $conn_start <= $end_time ) or
						( $conn_start >= $start_time and $conn_start <= $end_time ) )
					{
						$times{$conn_start} = $idx;
						$matching_lines[$idx] = $conn_struc;
						++$idx;
					}
				}
			}
		}
	}
	else
	{
		warn( "Failed to open incident temp file " . $RPT_CACHE->{incident}->{TEMP_FILE_NAME} . " for reading.\n" );
	}
	
	close( INFILE );
	
	# Sort the lines in ascending order
	foreach my $key( sort( {$a <=> $b} keys( %times ) ) )
	{
		push( @ret_strucs, $matching_lines[$times{$key}] );
	}
	
	return( @ret_strucs );
}


sub check_incident_struc
{
	foreach my $sus( keys( %{$RPT_CACHE->{incident}->{OFFENDERS}} ) )
	{
		if( ! $sus )
		{
			print "Suspect is undef\n";
			return( undef );
		}
		
		if( ! $RPT_CACHE->{incident}->{OFFENDERS}->{$sus} )
		{
			print "Incident data not defined\n";
			return( undef );
		}
		
		if( ! exists( $RPT_CACHE->{incident}->{OFFENDERS}->{$sus}->{VICTIMS} ) )
		{
			print "No data for VICTIMS\n";
			return( undef );
		}
		
		print "NEW OFFENDER: $sus\n";
		foreach my $vic( keys( %{$RPT_CACHE->{incident}->{OFFENDERS}->{$sus}->{VICTIMS}} ) )
		{
			if( ! $vic )
			{
				print "Victim is not defined\n";
				return( undef );
			}
			print "NEW VICTIM: $vic\n";
			
			if( ! $RPT_CACHE->{incident}->{OFFENDERS}->{$sus}->{VICTIMS}->{$vic} )
			{
				print "Victim data is not defined\n";
				return( undef );
			}
			
			foreach my $key( keys( %{$RPT_CACHE->{incident}->{OFFENDERS}->{$sus}->{VICTIMS}->{$vic}} ) )
			{
				print "KEY: $key\n";
			}
			
			foreach my $alarm( @{$RPT_CACHE->{incident}->{OFFENDERS}->{$sus}->{VICTIMS}->{$vic}->{ALARMS}} )
			{
				print "Message: ". Bro::Log::Alarm::message( $alarm ) ."\n";
			}
		}
		
		print "     END\n";
	}
}

sub availablereports
{
	my $sub_name = 'availablereports';
	
	my @ret_list = keys( %REPORT_MAP );
	
	return( @ret_list );
}

sub reportinputfunc
{
	my $sub_name = 'reportinputfunc';
	
	my $report_name = $_[0] || return( undef );
	
	if( exists( $REPORT_MAP{$report_name} ) )
	{
		return( $REPORT_MAP{$report_name}->{'input'} );
	}
	else
	{
		return( undef );
	}
}

sub reportoutputfunc
{
	my $sub_name = 'reportoutputfunc';
	
	my $report_name = $_[0] || return( undef );
	
	if( exists( $REPORT_MAP{$report_name} ) )
	{
		return( $REPORT_MAP{$report_name}->{'output'} );
	}
	else
	{
		return( undef );
	}
}

sub loadsignaturecode
{
	my $sub_name = 'loadsignaturecode';
	
	my @args = @_;
	my %match_sigs;
	my @signature_dirs;
	
	if( @args > 0 )
	{
		# If a list of signatures to find is given then only those will be 
		  # stored otherwise all signatures will be stored.
		
		foreach my $sigid( @args )
		{
			$match_sigs{$sigid} = 1;
		}
	}
	
	if( defined( $BRO_CONFIG->{BRO_SIG_DIR} ) or defined( $BRO_CONFIG->{BROPATH} ) )
	{
		my $rule_path;
		
		# if/else ladder is to allow for backwards compatability and will eventually
		# be removed.
		if( defined( $BRO_CONFIG->{BRO_SIG_DIR} ) and defined( $BRO_CONFIG->{BROPATH} ) )
		{
			$rule_path = join( ':', $BRO_CONFIG->{BRO_SIG_DIR}, $BRO_CONFIG->{BROPATH} );
		}
		elsif( defined( $BRO_CONFIG->{BROPATH} ) )
		{
			$rule_path = $BRO_CONFIG->{BROPATH};
		}
		else
		{
			$rule_path = $BRO_CONFIG->{BRO_SIG_DIR};
		}
		
		my $sigfile_suffix = $BRO_CONFIG->{BRO_SIG_SUFFIX};
		
		if( $rule_path =~ m/:/ )
		{
			@signature_dirs = split( /:/, $rule_path );
		}
		else
		{
			$signature_dirs[0] = $rule_path;
		}
		
		foreach my $sig_dir( @signature_dirs )
		{
			if( -d $sig_dir and -r $sig_dir and $sig_dir =~ m/^([[:print:]]+)$/ )
			{
				# Taint clean the directory name
				$sig_dir = $1;
				
				my @file_list;
				if( opendir( INDIR, $sig_dir ) )
				{
					foreach my $file_name( readdir( INDIR ) )
					{
						if( $file_name =~ m/^([[:print:]]+$sigfile_suffix)$/ )
						{
							# Taint clean the file name
							$file_name = $1;
							push( @file_list, $file_name );
						}
					}
				}
				else
				{
					warn( __PACKAGE__ . "::$sub_name, Unable to read directory $sig_dir\n" );
					next;
				}
				
				closedir( INDIR );
				
				foreach my $file_name( @file_list )
				{
					my $sig_file_data;
					my $full_file_path = join( '/', $sig_dir, $file_name );
					
					if( open( INFILE, $full_file_path ) )
					{
						local $/ = undef;
						$sig_file_data = <INFILE>;
					}
					else
					{
						warn( __PACKAGE__ . "::$sub_name, Unable to open signature file at $full_file_path\n" );
						next;
					}
					
					close( INFILE );
					
					foreach my $sig_block( findkeyblocks( $sig_file_data ) )
					{
						my $sig_obj = Bro::Signature->new( string => $sig_block );
						my $code_sigid = $sig_obj->sigid();
						if( !( %match_sigs ) or exists( $match_sigs{$code_sigid} ) )
						{
							$RPT_CACHE->{signaturecode}->{$code_sigid} = $sig_block;
						}
					}
				}
			}
		}
	}
	else
	{
		warn( __PACKAGE__ . "::$sub_name, neither BRO_SIG_DIR or BROPATH is defined\n" );
	}
}

sub signaturecode
{
	my $sub_name = 'signaturecode';
	
	my $sigid = $_[0] || return( undef );
	
	# Check on whether the signature block has been found and stored.
	if( $RPT_CACHE->{signaturecode}->{$sigid} )
	{
		return( $RPT_CACHE->{signaturecode}->{$sigid} );
	}
	else
	{
		return( '' );
	}
}


1;
