# Read and parse a Bro signature from a string or array reference
package Bro::Signature;

use strict;
require 5.006_001;
require Exporter;
use vars qw( $VERSION
			@ISA
			@EXPORT_OK
			$DEBUG );

# $Id: Signature.pm,v 1.4 2004/10/12 17:10:20 rwinslow Exp $
$VERSION = 1.14;
@EXPORT_OK = qw( findkeyblocks );
@ISA = qw( Exporter );
$DEBUG = 0;


sub new
{
	my $sub_name = 'new';

	my $self;
	my $proto = shift;
	my $class = ref( $proto ) || $proto;
	my %args = @_;
	my $sig_data = {};
	
	if( defined( $args{string} ) )
	{
		if( !( $sig_data = parsesig( $args{string} ) ) )
		{
			return( undef );
		}
	}
	else
	{
		warn( __PACKAGE__ . "::$sub_name, Signature must be passed in as a string for now.  Direct" .
			" object creation is not supported yet\n" );
		return( undef );
	}

	$self = $sig_data;
	bless( $self, $class );
	
	if( $DEBUG > 1 )
	{
		warn( $self->output() . "\n" );
	}
	return( $self );

}

sub parsesig
{
	my $sub_name = 'parsesig';

	my $sig_block = shift || return( undef );
	my $sig_data;
	my $ret_data = {};

	# Check on the storage of the data
	if( ref( $sig_block ) eq 'ARRAY' )
	{
		$sig_data = join( "\n", @{$sig_block} );
	}
	else
	{
		# Otherwise it gets treated as a string
		$sig_data = $sig_block;
	}

	my $parse_err = 0;

	if( $sig_data =~
		m/^signature[[:space:]]+([[:alnum:]_-]+)[[:space:]]*\{[[:space:]\n]+	# signature declaration
		(.+?)		# signature data
		[[:space:]]*\}$/xs )	# end of block
	{
		my $sig_id = $1;
		my $sig_options = $2;

		$ret_data->{sig_id} = $sig_id;

		my @raw_data;
		my $i = 0;

		foreach my $line( split( /\n/, $sig_options ) )
		{
			# Each line of the signature is stored in an
			 # array to maintain the order and any comments.
			 # The actual data is stored in a hash.
			 # Access of the data should be done through the methods
			 # to output the whole signature along with comments.

			# remove any leading spaces
			$line =~ s/^[[:space:]]*//;

			# Put the line into the raw_data array
			$raw_data[$i] = $line;

			# Strip any comments from the line
			$line =~ s/(?:[[:space:]]+|)#.+$//;

			my( $key, $value ) = split( /[[:space:]]+/, $line, 2 );
			# if there is still data in the line then process
			if( defined( $value ) )
			{
				# Remove any leading spaces
				$value =~ s/^[[:space:]]*//;
				# place a reference to the raw_data array for this attribute
				# set the value for this attribute instance
				push( @{$ret_data->{$key}}, { 
						'__raw_data_pos__' => $i,
						'__value__' => $value, },
				);
			}

			++$i;
		}

		$ret_data->{'__raw_data__'} = \@raw_data;
	}
	else
	{
		if( $DEBUG > 0 )
		{
			warn( __PACKAGE__ . "::$sub_name, Signature read error. Could not find a valid signature block\n" );
		}
		$parse_err = 1;
	}

	if( $parse_err )
	{
		return( undef );
	}
	else
	{
		return( $ret_data );
	}
}

sub findkeyblocks
{
	my $sub_name = 'findkeyblocks';

	my $string = $_[0];
	my $keyword = 'signature';
	my $key_len = length( $keyword );
	my @ret_blocks;
	my $block_idx = 0;			# Current idx of the keyword blocks that will be returned
	my $open_brace_count = 0;	# Number of open braces encountered
	my $close_brace_count = 0;	# number of close braces encountered
	my $len = length( $string );	# length of string passed in
	my $pos_idx = 0;			# current position of substring
	my $st_blk_pos = 0;			# string position of a new block
	my $blk_len = 0;			# length of code block
	my $found_key_start = 0;		# flag that a new keyword block has started
	my $comment_line = 0;		# flag if comments are in effect
	my $quoted;			# flag if quoting is active. contents are the quote character

	# make sure the string is of non-zero length.  Add one more so our
	 # loop below works
	if( $len > 0 )
	{
		++$len;
	}
	else
	{
		return( undef );
	}

	# Push a newline onto the front of the string.  This just helps in
	 # starting the pattern matching.  It's easier than writting a
	 # bunch more lines of code.
	$string = "\n" . $string;
	++$len;

	while( $pos_idx < $len )
	{
		# If the two numbers match then a complete block has been found.
		if( $open_brace_count == $close_brace_count and
			$open_brace_count > 0 )
		{
			$ret_blocks[$block_idx] = substr( $string, $st_blk_pos, $blk_len );
			
			if( $DEBUG > 2 )
			{
				my $end_pos = $st_blk_pos + $blk_len;
				warn( __PACKAGE__ . "::$sub_name, Found signature block between bytes $st_blk_pos and $end_pos\n" );
				warn( $ret_blocks[$block_idx] . "\n" );
			}
			
			++$block_idx;
			$open_brace_count = 0;
			$close_brace_count = 0;
			$found_key_start = 0;
			$st_blk_pos = $pos_idx;
			$blk_len = 0;
		}
		elsif( $close_brace_count > $open_brace_count )
		{
			# Looks like there was a syntax error perhaps the
			 # file contains a keyword but no valid block
			 # Reset the counters and backup to $st_blk_pos + 1
			if( $DEBUG > 0 )
			{
				warn( __PACKAGE__ . "::$sub_name, Error in parsing $keyword, found closed brace",
					" for a block but no open brace at char position $pos_idx\n" );
			}
			$open_brace_count = 0;
			$close_brace_count = 0;
			$found_key_start = 0;
			$pos_idx = $st_blk_pos + 1;
			$st_blk_pos = '';
			$blk_len = 0;
			next;
		}

		# Make sure that substr returns a value
		if( defined( my $wc = substr( $string, $pos_idx, 1 ) ) )
		{
			if( $quoted
				and $found_key_start )
			{
				if( $wc eq $quoted
					and substr( $string, $pos_idx - 1, 1 ) ne '\\' )
				{
					$quoted = undef;
				}
			}
			# check for comment '#' character which is active until the 
			 # start of a newline
			elsif( ! $comment_line and 
				$wc eq '#' and
				substr( $string, $pos_idx - 1, 1 ) ne '\\' )
			{
				$comment_line = 1;
			}
			# Check if the comment flag is set
			elsif( $comment_line )
			{
				# If the current character is a newline then reset the comment flag
				if( $wc =~ m/\n/ )
				{
					$comment_line = 0;
				}
			}
			# start this if not currently working in a block and not a comment
			elsif( ! $found_key_start and ! $comment_line )
			{
				# check if the keyword is found 
				if( $pos_idx + $key_len + 1 < $len
				and substr( $string, $pos_idx, $key_len ) eq $keyword )
				{
					# check to make sure that the keyword is followed by a space
					if( substr( $string, $pos_idx + $key_len, 1 ) =~
						m/[[:space:]]/ )
					{
						$found_key_start = 1;
						$st_blk_pos = $pos_idx;
					}
				}
			}
		# Need to re-think this one at some point though it's not going to kill things
		  # to leave it out right now.
		
		# Check if a nested block has started.
		#	elsif( $pos_idx + $key_len + 1 < $len
		#		and $found_key_start
		#		and substr( $string, $pos_idx, $key_len ) eq $keyword )
		#	{
		#		if( substr( $string, $pos_idx + $key_len, 1 ) =~
		#				m/[[:space:]]/ )
		#		{
		#			if( $DEBUG > 0 )
		#			{
		#				warn( __PACKAGE__ . "::$sub_name, New $keyword keyword found inside of another $keyword block",
		#					" at char position $pos_idx\n" );
		#
		#				#print "STRING => $string\n";
		#			}
		#
		#			# Reset the search params
		#			$open_brace_count = 0;
		#			$close_brace_count = 0;
		#			$found_key_start = 0;
		#			++$pos_idx;
		#			$st_blk_pos = '';
		#			$blk_len = 0;
		#			next;
		#		}
		#	}
			elsif( $wc eq '{' and 
				substr( $string, $pos_idx - 1, 1 ) ne '\\'
				and substr( $string, $pos_idx + 1, 1 ) =~ m/[[:space:]\n]/ )
			{
				++$open_brace_count;
			}
			elsif( $wc eq '}' and
				substr( $string, $pos_idx - 1, 1 ) ne '\\'
				and substr( $string, $pos_idx - 1, 1 ) =~ m/[[:space:]\n]/ )
			{
				++$close_brace_count;
			}
			elsif( ! $quoted
				and ! $comment_line
				and $found_key_start )
			{
				if( $wc eq '"'
					and substr( $string, $pos_idx - 1, 1 ) ne '\\' )
				{
					$quoted = $wc;
				}
			}
			else
			{

			}

			# Only append chars to the working string if in a $keyword block
			if( $found_key_start )
			{
				++$blk_len;
			}
			++$pos_idx;
		}
		else
		{
			warn( __PACKAGE__ . "::$sub_name, Failed to pull data out using substr at position $pos_idx\n" );
			++$pos_idx;
		}
	}

	if( wantarray )
	{
		return( @ret_blocks );
	}
	else
	{
		return( \@ret_blocks );
	}
}

sub addcomment
{
	my $sub_name = 'addcomment';

	my $self = shift || return( undef );
	my $comment = shift || return( undef );

	# Make sure the comment starts with a '#'
	if( $comment !~ m/^[[:space:]]*#/ )
	{
		$comment = '#' . $comment;
	}

	if( $self->{'__raw_data__'} )
	{
		my $next_idx = $#{$self->{'__raw_data__'}} + 1;
		$self->{'__raw_data__'}->[$next_idx] = $comment;
	}
	else
	{
		return( undef );
	}

	return( 1 );
}

sub addoption
{
	my $sub_name = 'addoption';

	my $self = shift || return( undef );
	my $option = shift || return( undef );
	my $option_data = shift || return( undef );

	if( $self->{'__raw_data__'} )
	{
		my $next_idx = $#{$self->{'__raw_data__'}} + 1;
		$self->{'__raw_data__'}->[$next_idx] = $option . ' ' . $option_data;
		push( @{$self->{$option}}, 
			{ '__raw_data_pos__' => $next_idx,
			'__value__' => $option_data, } );
	}
	else
	{
		return( undef );
	}

	return( 1 );
}

sub deloption
{
	my $sub_name = 'deloption';
	# Accepts at a minimum the object and an option to remove
	# For options that have multivalues the value can be passed in as 
	 # option_data to remove only the one value from the object.
	# If an option has multivalues and no option_data is given then all
	 # options are removed.


	my $self = shift || return( undef );
	my $match_option = shift || return( undef );
	my $match_data = shift;	#optional
	my $success = 0;
	my @raw_data_pos;

	if( defined( $match_data ) )
	{
		if( $DEBUG > 4 )
		{
			warn( __PACKAGE__ . "::$sub_name, Method $sub_name has been asked to delete option '" .
				$match_option . "', value '" . $match_data . "'\n" );
		}
		# Must match on both the option key and it's value
		if( exists( $self->{$match_option} ) )
		{				
			if( ref( $self->{$match_option} ) eq 'ARRAY' )
			{
				if( $DEBUG > 4 )
				{
					warn( __PACKAGE__ . "::$sub_name, Found a match on $match_option in signature\n" );
				}

				my $__found = 0;
				my @good_data;
				foreach my $opt_value( @{$self->{$match_option}} )
				{
					if( $opt_value->{'__value__'} eq $match_data )
					{
						++$__found;
						$success = 1;
						push( @raw_data_pos, $opt_value->{'__raw_data_pos__'} );
					}
					elsif( defined( my $tt = $opt_value->{'__value__'} ) )
					{
						push( @good_data, $opt_value );
					}

				}

				if( $success )
				{
					if( defined( $good_data[0] ) )
					{
						# Replace the object's old data with the good data
						$self->{$match_option} = [ @good_data ];
					}
					else
					{
						delete $self->{$match_option};
					}
				}
			}
		}
	}
	elsif( exists( $self->{$match_option} ) )
	{
		if( $DEBUG > 4 )
		{
			warn( __PACKAGE__ . "::$sub_name, Method $sub_name has been asked to delete all data with" .
				" an option name of $match_option\n" );
		}

		foreach my $opt_data( @{$self->{$match_option}} )
		{
			push( @raw_data_pos, $opt_data->{'__raw_data_pos__'} );
		}
		delete $self->{$match_option};
		$success = 1;
	}


	if( $success )
	{
		# cleanup the __raw_data__ storage in the object
		# get the last index of the __raw_data__ array
		foreach my $idx( @raw_data_pos )
		{
			$self->{'__raw_data__'}->[$idx] = undef;
		}

		# Delete the 
	}

	return( $success );

}

sub output
{
	my $sub_name = 'output';

	my $self = shift || return( undef );
	my %args = @_;
	my $ret_string;
	my $prefix = '';
	my $comments = 0;

	# Check on options
	if( $args{sigprefix} )
	{
		$prefix = $args{sigprefix};
	}

	if( defined( $args{comments} ) )
	{
		$comments = $args{comments};
	}
	else
	{
		$comments = 1;
	}

	# opening to a Bro signature block
	$ret_string = 'signature ' . $prefix . $self->{sig_id} . ' {' . "\n";

	if( $comments )
	{
		foreach my $line( @{$self->{'__raw_data__'}} )
		{
			if( defined( $line ) )
			{
				$ret_string = $ret_string . '  ' . $line . "\n";
			}
		}
	}
	else
	{
		while( my( $option, $opt_array ) = each( %{$self} ) )
		{
			if( $option ne '__raw_data__' and $option ne 'sig_id' )
			{
				foreach my $opt_data( @{$opt_array} )
				{
					$ret_string = $ret_string . 
					'  ' .
					$option .
					' ' .
					$opt_data->{'__value__'} .
					"\n";
				}
			}
		}
	}

	# closing of the Bro signature block
	$ret_string = $ret_string . '}';

	return( $ret_string );
}

sub option
{
	my $sub_name = 'option';

	my $self = shift || return( undef );
	my $sig_option = shift || return( undef );
	my $ret_data;

	if( $self->{$sig_option} )
	{
		foreach my $data( @{$self->{$sig_option}} )
		{
			push( @{$ret_data}, $data->{'__value__'} );
		}
	}
	else
	{
		return( undef );
	}

	if( @{$ret_data} > 1 )
	{
		return( @{$ret_data} );
	}
	else
	{
		return( $ret_data->[0] );
	}
}

sub sigid
{
	# This is the entire Bro signature id as parsed in from a 
	 # signature block
	my $sub_name = 'sigid';

	my $self = shift || return( undef );

	if( $self->{sig_id} )
	{
		return( $self->{sig_id} );
	}
	else
	{
		return( undef );
	}
}

1;