// $Id: File.cc,v 1.2 2004/09/17 03:52:27 vern Exp $
//
// Copyright (c) 1996, 1997, 1998, 1999, 2002
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "config.h"

#include <sys/types.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/resource.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "File.h"
#include "Type.h"
#include "Timer.h"
#include "Expr.h"
#include "NetVar.h"
#include "Serializer.h"

// Timer which on dispatching rotates the file.
class RotateTimer : public Timer {
public:
	RotateTimer(double t, BroFile* f) : Timer(t, TIMER_ROTATE)	{ file = f; }

	virtual void Dispatch(double t, int is_expire)
		{
		if ( active && ! is_expire )
			file->Rotate();
		}

protected:
	BroFile* file;
};


// The following could in principle be part of a "file manager" object.

#define MAX_OPEN_FILE_CACHE_SIZE 32
static int num_open_files = 0;
static int max_open_files = 0;
static BroFile* head = 0;
static BroFile* tail = 0;


// Maximizes the number of open file descriptors and returns the number
// that we should use for the cache.
static int maximize_num_fds()
	{
#ifdef NO_HAVE_SETRLIMIT
	return MAX_OPEN_FILE_CACHE_SIZE;
#else
	struct rlimit rl;
	if ( getrlimit(RLIMIT_NOFILE, &rl) < 0 )
		internal_error("maximize_num_fds(): getrlimit failed");

	if ( rl.rlim_max == RLIM_INFINITY )
		{
		// Don't try raising the current limit.
		if ( rl.rlim_cur == RLIM_INFINITY )
			// Let's not be too ambitious.
			return MAX_OPEN_FILE_CACHE_SIZE;
		else
			return rl.rlim_cur / 2;
		}

	// See if we can raise the current to the maximum.
	rl.rlim_cur = rl.rlim_max;

	if ( setrlimit(RLIMIT_NOFILE, &rl) < 0 )
		internal_error("maximize_num_fds(): setrlimit failed");

	return rl.rlim_cur / 2;
#endif
	}


BroFile::BroFile(FILE* arg_f)
	{
	Init();
	f = arg_f;
	name = access = 0;
	t = base_type(TYPE_STRING);
	is_open = (f != 0);
	}

BroFile::BroFile(FILE* arg_f, const char* arg_name, const char* arg_access)
	{
	Init();
	f = arg_f;
	name = copy_string(arg_name);
	access = copy_string(arg_access);
	t = base_type(TYPE_STRING);
	is_open = (f != 0);
	}

BroFile::BroFile(const char* arg_name, const char* arg_access, BroType* arg_t)
	{
	Init();

	name = copy_string(arg_name);
	access = copy_string(arg_access);
	t = arg_t ? arg_t : base_type(TYPE_STRING);
	if ( ! Open() )
		{
		error(fmt("cannot open %s: %s", name, strerror(errno)));
		exit(1);
		}
	}

bool BroFile::Open()
	{
	open_time = (time_t) network_time;

	if ( ! max_open_files )
		// Haven't initialized yet.
		max_open_files = maximize_num_fds();

	if ( num_open_files >= max_open_files )
		PurgeCache();

	f = fopen(name, access);

	if ( f )
		{
		InsertAtBeginning();
		is_managed = 1;
		}
	else
		{
		// No point managing it.
		is_open = is_managed = 0;
		return false;
		}

	return true;
	}

BroFile::~BroFile()
	{
	Close();
	Unref(t);
	Unref(attrs);

	delete [] name;
	delete [] postproc;
	delete [] access;

#ifdef USE_OPENSSL
	delete [] cipher_buffer;
#endif
	}

void BroFile::Init()
	{
	is_managed = 0;
	position = 0;
	next = prev = 0;
	postproc = 0;
	rotate = 0;
	rotate_timer = 0;
	open_time = 0;
	attrs = 0;

#ifdef USE_OPENSSL
	pub_key = 0;
	cipher_ctx = 0;
	cipher_buffer = 0;
#endif
	}

FILE* BroFile::File()
	{
	if ( is_managed )
		{
		if ( is_open )
			MoveToBeginning();
		else
			{
			if ( f )
				internal_error("BroFile non-nil non-open file");

			if ( num_open_files >= max_open_files )
				PurgeCache();

			if ( position == 0 )
				// Need to truncate it.
				f = fopen(name, access);
			else
				// Don't clobber it.
				f = fopen(name, "a");

			if ( ! f )
				{
				f = fopen("/dev/null", "w");

				if ( ! f )
					internal_error("out of file descriptors");
				run_time("can't open %s", this);
				is_managed = 0;
				return f;
				}

			if ( fseek(f, position, SEEK_SET) < 0 )
				run_time("reopen seek failed", this);

			InsertAtBeginning();
			}
		}

	return f;
	}

FILE* BroFile::Seek(long new_position)
	{
	if ( ! File() )
		return 0;

	if ( fseek(f, new_position, SEEK_SET) < 0 )
		run_time("seek failed", this);

	return f;
	}

void BroFile::SetBuf(bool buffered)
	{
	if ( ! f )
		return;

	if ( setvbuf(f, NULL, buffered ? _IOFBF : _IOLBF, 0) != 0 )
		run_time("setvbuf failed", this);
	}

int BroFile::Close()
	{
#ifdef USE_OPENSSL
	FinishEncrypt();
#endif

	// Do not close stdout/stderr.
	if ( f == stdout || f == stderr )
		return 0;

	if ( is_managed )
		{
		Unlink();
		if ( f )
			{
			fclose(f);
			f = 0;
			Postprocess(name);
			open_time = 0;
			}

		is_managed = 0;	// no longer managed since will never reopen

		return 1;
		}

	// Not managed.
	if ( ! f )
		return 0;

	fclose(f);
	f = 0;

	return 1;
	}

void BroFile::Suspend()
	{
	if ( ! is_managed )
		internal_error("BroFile::Suspend() called for non-managed file");
	if ( ! is_open )
		internal_error("BroFile::Suspend() called for non-open file");
	Unlink();

	if ( ! f )
		internal_error("BroFile::Suspend() called for nil file");

	if ( (position = ftell(f)) < 0 )
		{
		run_time("ftell failed", this);
		position = 0;
		}

	fclose(f);
	f = 0;
	}

void BroFile::PurgeCache()
	{
	if ( ! tail )
		internal_error("BroFile purge of empty cache");

	tail->Suspend();
	}

void BroFile::Unlink()
	{
	if ( is_open )
		{
		if ( head == this )
			head = Next();
		else
			Prev()->SetNext(next);

		if ( tail == this )
			tail = Prev();
		else
			Next()->SetPrev(prev);

		if ( (head || tail) && ! (head && tail) )
			internal_error("BroFile link list botch");

		is_open = 0;
		prev = next = 0;

		if ( --num_open_files < 0 )
			internal_error("BroFile open files underflow");
		}
	}

void BroFile::InsertAtBeginning()
	{
	if ( ! head )
		{
		head = tail = this;
		next = prev = 0;
		}
	else
		{
		SetNext(head);
		SetPrev(0);
		head->SetPrev(this);
		head = this;
		}

	is_open = 1;

	if ( ++num_open_files > max_open_files )
		internal_error("BroFile open files overflow");
	}

void BroFile::MoveToBeginning()
	{
	if ( head == this )
		return;	// already at the beginning

	if ( ! is_open || ! prev )
		internal_error("BroFile inconsistency in MoveToBeginning()");

	Unlink();
	InsertAtBeginning();
	}

void BroFile::Describe(ODesc* d) const
	{
	d->AddSP("file");

	if ( name )
		{
		d->Add("\"");
		d->Add(name);
		d->AddSP("\"");
		}

	d->AddSP("of");
	if ( t )
		t->Describe(d);
	else
		d->Add("(no type)");
	}

void BroFile::SetAttrs(Attributes* arg_attrs)
	{
	if ( ! arg_attrs )
		return;

	attrs = arg_attrs;
	Ref(attrs);

	Attr* ef = attrs->FindAttr(ATTR_ROTATE);
	if ( ef )
		rotate = ef->AttrExpr()->ExprVal()->AsInterval();

	ef = attrs->FindAttr(ATTR_POSTPROCESSOR);
	if ( ef )
		{
		delete [] postproc;
		postproc = copy_string(ef->AttrExpr()->ExprVal()->AsString()->CheckString());
		}

	ef = attrs->FindAttr(ATTR_ENCRYPT);
	if ( ef )
		{
		if ( ef->AttrExpr() )
			InitEncrypt(ef->AttrExpr()->ExprVal()->AsString()->CheckString());
		else
			InitEncrypt(log_encryption_key->AsString()->CheckString());
		}

	InstallRotateTimer();
	}

void BroFile::Rotate()
	{
	if ( ! is_open )
		return;

	if ( dont_rotate )
		{
		open_time = (time_t) network_time;
		InstallRotateTimer();
		return;
		}

	// Build file names.
	const int buflen = strlen(name) + 128;

	char tmpname[buflen], newname[buflen+4];

	snprintf(newname, buflen, "%s.%d.%d", name, getpid(), int(time(0)));
	newname[buflen-1] = '\0';
	strcpy(tmpname, newname);
	strcat(tmpname, ".tmp");

	// First open the new file using a temporary name.
	FILE* newf = fopen(tmpname, "w");
	if ( ! newf )
		{
		run_time(fmt("rotation: can't open %s: %s", tmpname, strerror(errno)));
		InstallRotateTimer();
		return;
		}

	// Then move old file to "<name>.<pid>.<timestamp>" and make sure
	// it really gets created.
	struct stat dummy;
	if ( link(name, newname) < 0 || stat(newname, &dummy) < 0 )
		{
		run_time(fmt("rotation: can't move %s to %s: %s", name, newname, strerror(errno)));

		fclose(newf);
		unlink(newname);
		unlink(tmpname);

		InstallRotateTimer();

		return;
		}

	// Close current file, and move the tmp to its place
	if ( unlink(name) < 0 || link(tmpname, name) < 0 || unlink(tmpname) < 0 )
		{
		run_time(fmt("can't move %s to %s: %s", tmpname, name, strerror(errno)));
		exit(1); // Hard to fix, but shouldn't really happen anyway...
		}

	fclose(f);
	Postprocess(newname);

	open_time = (time_t) network_time;
	f = newf;

	InstallRotateTimer();
	}

void BroFile::Postprocess(const char* file)
	{
	if ( ! (postproc && *postproc && open_time) )
		return;

	pid_t pid = fork();
	if ( pid < 0 )
		{
		run_time("can't fork postprocessor");
		return;
		}

	if ( pid == 0 )
		{
		// Child.

		// Redirect stdin/stdout/stderr to /dev/null.
		int fd = open("/dev/null", O_RDWR);
		if ( fd < 0 )
			{
			fprintf(stderr, "can't redirect file descriptors");
			_exit(1);
			}

		close(0);
		close(1);
		close(2);

		dup2(fd, 0);
		dup2(fd, 1);
		dup2(fd, 2);

		close(fd);

		// Be nice ...
		setpriority(PRIO_PROCESS, 0, 10);

		// Build arguments.
		char opened[64];
		char closed[64];

		time_t current = (time_t) network_time;
		strftime(closed, sizeof(closed),
				"%Y-%m-%d-%H-%M-%S", localtime(&current));
		strftime(opened, sizeof(opened),
				"%Y-%m-%d-%H-%M-%S", localtime(&open_time));

		// Paranoia.
		opened[sizeof(opened) - 1] = closed[sizeof(closed) - 1] = '\0';

		// Run child.
		execlp(postproc, postproc, file, name, opened, closed, 0);

		// If reached this point, we got an error.
		run_time(fmt("can't run postprocessor: %s", strerror(errno)));
		_exit(1);
		}

	// Parent.
	return;
	}

void BroFile::InstallRotateTimer()
	{
	if ( rotate )
		{
		// When this is called for the first time, network_time can
		// still be zero. If so, we set a timer which fires
		// immediately but doesn't rotate when it expires.
		if ( ! network_time )
			{
			rotate_timer = new RotateTimer(1, this);
			dont_rotate = true;
			}

		else
			{
			// To get nice timestamps, we round the time up to
			// the next multiple of the rotation interval.
			double t = floor(network_time / rotate) * rotate +
					rotate;
			rotate_timer = new RotateTimer(t, this);
			dont_rotate = false;
			}

		timer_mgr->Add(rotate_timer);
		}
	}


void BroFile::SetDefaultRotation(double interval, const char* arg_postproc)
	{
	for ( BroFile* f = head; f; f = f->next )
		{
		if ( ! (f->attrs && f->attrs->FindAttr(ATTR_ROTATE)) )
			{
			f->rotate = interval;
			f->InstallRotateTimer();
			}

		if ( ! (f->attrs && f->attrs->FindAttr(ATTR_POSTPROCESSOR)) )
			f->postproc = copy_string(arg_postproc);
		}
	}

void BroFile::CloseManagedFiles()
	{
	BroFile* next;
	for ( BroFile* f = head; f; f = next )
		{
		next = f->next;
		if ( f->is_managed )
			f->Close();
		}
	}

#ifndef USE_OPENSSL

void BroFile::InitEncrypt(const char* keyfile)
	{
	if ( keyfile )
		{
		error("file encryption requested, but OpenSSL support not compiled in.");
		Close();
		}
	}

#else

void BroFile::InitEncrypt(const char* keyfile)
	{
	if ( ! (pub_key || keyfile) )
		return;

	if ( ! pub_key )
		{
		FILE* key = fopen(keyfile, "r");

		if ( ! key )
			{
			error(fmt("can't open key file %s: %s", keyfile, strerror(errno)));
			Close();
			return;
			}

		pub_key = PEM_read_PUBKEY(key, 0, 0, 0);
		if ( ! pub_key )
			{
			error(fmt("can't read key from %s: %s", keyfile,
					ERR_error_string(ERR_get_error(), 0)));
			Close();
			return;
			}
		}

	// Depending on the OpenSSL version, EVP_*_cbc()
	// returns a const or a non-const.
	EVP_CIPHER* cipher_type = (EVP_CIPHER*) EVP_bf_cbc();
	cipher_ctx = new EVP_CIPHER_CTX;

	unsigned char secret[EVP_PKEY_size(pub_key)];
	unsigned char* psecret = secret;
	unsigned long secret_len;

	int iv_len = EVP_CIPHER_iv_length(cipher_type);
	unsigned char iv[iv_len];

	if ( ! EVP_SealInit(cipher_ctx, cipher_type, &psecret,
				(int*) &secret_len, iv, &pub_key, 1) )
		{
		error(fmt("can't init cipher context for %s: %s", keyfile,
				ERR_error_string(ERR_get_error(), 0)));
		Close();
		return;
		}

	secret_len = htonl(secret_len);

	if ( ! (fwrite("BROENC1", 7, 1, f) &&
		fwrite(&secret_len, sizeof(secret_len), 1, f) &&
		fwrite(secret, ntohl(secret_len), 1, f) &&
		fwrite(iv, iv_len, 1, f)) )
		{
		error(fmt("can't write header to log file %s: %s",
				name, strerror(errno)));
		Close();
		return;
		}

	int buf_size = MIN_BUFFER_SIZE + EVP_CIPHER_block_size(cipher_type);
	cipher_buffer = new unsigned char[buf_size];
	}
#endif

void BroFile::FinishEncrypt()
	{
#ifdef USE_OPENSSL
	if ( ! pub_key )
		return;

	if ( cipher_ctx )
		{
		int outl;
		EVP_SealFinal(cipher_ctx, cipher_buffer, &outl);

		if ( outl && ! fwrite(cipher_buffer, outl, 1, f) )
			{
			run_time(fmt("write error for %s: %s",
					name, strerror(errno)));
			return;
			}

		delete cipher_ctx;
		cipher_ctx = 0;
		}
#endif
	}


int BroFile::Write(const char* data, int len)
	{
	if ( ! is_open )
		return 0;

	if ( ! len )
		len = strlen(data);

#ifdef USE_OPENSSL
	if ( cipher_ctx )
		{
		while ( len )
			{
			int outl;
			int inl = min(MIN_BUFFER_SIZE, len);

			if ( ! EVP_SealUpdate(cipher_ctx, cipher_buffer, &outl,
						(unsigned char*)data, inl) )
				{
				run_time(fmt("encryption error for %s: %s",
					name,
					ERR_error_string(ERR_get_error(), 0)));
				Close();
				return 0;
				}

			if ( outl && ! fwrite(cipher_buffer, outl, 1, f) )
				{
				run_time(fmt("write error for %s: %s",
						name, strerror(errno)));
				Close();
				return 0;
				}

			data += inl;
			len -= inl;
			}

		return 1;
		}
#endif
	return fwrite(data, len, 1, f) > 0;
	}

bool BroFile::Serialize(SerialInfo* info) const
	{
	return SerialObj::Serialize(info);
	}

BroFile* BroFile::Unserialize(UnserialInfo* info)
	{
	BroFile* file = (BroFile*) SerialObj::Unserialize(info, SER_BRO_FILE);

	if ( file->is_open )
		return file;

	// If there is already an object for this file, return it.
	if ( file->name )
		{
		for ( BroFile* f = head; f; f = f->next )
			{
			if ( f->name && streq(file->name, f->name) )
				{
				Unref(file);
				Ref(f);
				return f;
				}
			}
		}

	// Otherwise, open.
	if ( ! file->Open() )
		{
		info->s->Error(fmt("cannot open %s: %s",
					file->name, strerror(errno)));
		return 0;
		}

	return file;
	}

IMPLEMENT_SERIAL(BroFile, SER_BRO_FILE);

bool BroFile::DoSerialize(SerialInfo* info) const
	{
	DO_SERIALIZE(SER_BRO_FILE, BroObj);

	const char* s = name;

	if ( ! is_managed )
		{
		// We can handle stdin/stdout/stderr but no others.
		if ( f == stdin )
			s = "/dev/stdin";
		else if ( f == stdout )
			s = "/dev/stdout";
		else if ( f == stderr )
			s = "/dev/stderr";
		else
			{
			info->s->Error("can't serialize files we don't manage");
			return false;
			}
		}

	if ( ! (SERIALIZE(s) && SERIALIZE(access) && t->Serialize(info)) )
		return false;

	SERIALIZE_OPTIONAL(attrs);
	return true;
	}

bool BroFile::DoUnserialize(UnserialInfo* info)
	{
	DO_UNSERIALIZE(BroObj);

	if ( ! UNSERIALIZE_STR(&name, 0) || ! UNSERIALIZE_STR(&access, 0) )
		return false;

	t = BroType::Unserialize(info);
	if ( ! t )
		return false;

	UNSERIALIZE_OPTIONAL(attrs, Attributes::Unserialize(info));

	// Parse attributes.
	SetAttrs(attrs);
	// SetAttrs() has ref'ed attrs again.
	Unref(attrs);

	// Bind stdin/stdout/stderr.
	FILE* file = 0;
	is_open = false;
	f = 0;

	if ( streq(name, "/dev/stdin") )
		file = stdin;
	else if ( streq(name, "/dev/stdout") )
		file = stdout;
	else if ( streq(name, "/dev/stderr") )
		file = stderr;

	if ( file )
		{
		delete [] name;
		name = 0;
		f = file;
		is_open = true;
		}

	return true;
	}



