// $Id: Net.cc,v 1.5 2004/09/20 06:13:57 vern Exp $
//
// Copyright (c) 1996, 1997, 1998, 1999, 2000, 2001, 2002
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "config.h"

#include <sys/types.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>

#include "NetVar.h"
#include "Sessions.h"
#include "Event.h"
#include "Timer.h"
#include "Var.h"
#include "Logger.h"
#include "Net.h"
#include "TCP_Rewriter.h"
#include "Anon.h"
#include "PacketSort.h"
#include "Serializer.h"
#include "RemoteSerializer.h"

extern "C" {
#include "setsignal.h"
};

extern "C" {
extern int select(int, fd_set *, fd_set *, fd_set *, struct timeval *);
}

PList(PktSrc) pkt_srcs;

pcap_dumper_t* pkt_dump = 0;
// For trace of rewritten packets
PacketDumper* transformed_pkt_dump = 0;
// For trace of original packets from selected connections
PacketDumper* source_pkt_dump = 0;
int transformed_pkt_dump_MTU = 1514;

int reading_live = 1;	// 1 if reading from network interface, 0 if save file
char* user_pcap_filter = 0;

double network_time = 0.0;	// time according to last packet timestamp
double processing_start_time = 0.0;	// time started working on current pkt
double last_watchdog_proc_time = 0.0;	// value of above during last watchdog
static double next_packet_time = 0.0;	// earliest time next packet may arrive
const double max_packet_time = 1e30;
#define sort_packets()	(packet_sort_window > 0)

static const struct pcap_pkthdr* current_hdr = 0;
static const u_char* current_pkt = 0;
static int current_dispatched = 0;
static PktSrc* current_pktsrc = 0;

RETSIGTYPE watchdog(int /* signo */)
	{
	if ( processing_start_time != 0.0 )
		{
		// The signal arrived while we're processing a packet and/or
		// its corresponding event queue.  Check whether we've been
		// spending too much time, which we take to mean we've wedged.

		// Note that it's subtle how exactly to test this.  In
		// processing_start_time we have the timestamp of the packet
		// we're currently working on.  But that *doesn't* mean that
		// we began work on the packet at that time; we could have
		// begun at a much later time, depending on how long the
		// packet filter waited (to fill its buffer) before handing
		// up this packet.  So what we require is that the current
		// processing_start_time matches the processing_start_time we
		// observed last time the watchdog went off.  If so, then
		// we've been working on the current packet for at least
		// watchdog_interval seconds.

		if ( processing_start_time == last_watchdog_proc_time )
			{
			// snprintf() calls alloc/free routines if you use %f!
			// We need to avoid doing that given we're in a single
			// handler and the allocation routines are not
			// reentrant.

			double ct = current_time();

			int int_ct = int(ct);
			int frac_ct = int((ct - int_ct) * 1e6);

			int int_pst = int(processing_start_time);
			int frac_pst =
				int((processing_start_time - int_pst) * 1e6);

			char msg[512];
			snprintf(msg, sizeof(msg),
				"**watchdog timer expired, t = %d.%06d, start = %d.%06d, dispatched = %d",
				int_ct, frac_ct, int_pst, frac_pst,
				current_dispatched);

			bro_logger->Log(msg);
			run_time("watchdog timer expired");

			if ( current_hdr )
				pcap_dump((u_char*) pkt_dump, current_hdr, current_pkt);
			net_get_final_stats();
			net_finish(0);

			abort();
			exit(1);
			}
		}

	last_watchdog_proc_time = processing_start_time;

	(void) alarm(watchdog_interval);
	return RETSIGVAL;
	}

void net_init(name_list& interfaces, name_list& readfiles,
	        const char* writefile, const char* transformed_writefile,
	        const char* filter, const char* secondary_filter,
		int do_watchdog)
	{
	init_net_var();

	if ( readfiles.length() > 0 )
		{
		reading_live = 0;

		for ( int i = 0; i < readfiles.length(); ++i )
			{
			PktFileSrc* ps = new PktFileSrc(readfiles[i], filter);

			if ( ps->NotOpen() )
				{
				fprintf(stderr, "%s: problem with trace file %s - %s\n",
					prog, readfiles[i], ps->ErrorMsg());
				exit(1);
				}
			else
				pkt_srcs.append(ps);

			if ( secondary_filter )
				{
				// We use a second PktFileSrc for the
				// secondary path.
				PktFileSrc* ps = new PktFileSrc(readfiles[i],
							secondary_filter,
							TYPE_FILTER_SECONDARY);

				if ( ps->NotOpen() )
					{
					fprintf(stderr, "%s: problem with trace file %s - %s\n",
						prog, readfiles[i], ps->ErrorMsg());
					exit(1);
					}
				else
					pkt_srcs.append(ps);

				ps->AddSecondaryTablePrograms();
				}
			}
		}

	else if ( interfaces.length() > 0 )
		{
		reading_live = 1;

		for ( int i = 0; i < interfaces.length(); ++i )
			{
			PktInterfaceSrc* ps =
				new PktInterfaceSrc(interfaces[i], filter);

			if ( ps->NotOpen() )
				{
				fprintf(stderr, "%s: problem with interface %s - %s\n",
					prog, interfaces[i], ps->ErrorMsg());
				exit(1);
				}
			else
				pkt_srcs.append(ps);

			if ( secondary_filter )
				{
				PktInterfaceSrc* ps =
					new PktInterfaceSrc(interfaces[i],
						filter, TYPE_FILTER_SECONDARY);

				if ( ps->NotOpen() )
					{
					fprintf(stderr, "%s: problem with interface %s - %s\n",
						prog, interfaces[i], ps->ErrorMsg());
					exit(1);
					}
				else
					pkt_srcs.append(ps);

				ps->AddSecondaryTablePrograms();
				}
			}
		}

	if ( writefile )
		{
		// ### This will fail horribly if there are multiple
		// interfaces with different-lengthed media.
		pkt_dump = pkt_srcs[0]->OpenDumper(writefile);
		if ( ! pkt_dump )
			{
			fprintf(stderr, "%s: can't open write file \"%s\" - %s\n",
				prog, writefile, pkt_srcs[0]->ErrorMsg());
			exit(1);
			}
		}

	if ( transformed_writefile )
		{
		pcap_dumper_t* dumper = pkt_srcs[0]->OpenDumper(transformed_writefile);
		if ( ! dumper )
			{
			fprintf(stderr, "%s: can't open trace transformation write file \"%s\" - %s\n",
				prog, writefile, pkt_srcs[0]->ErrorMsg());
			exit(1);
			}

		transformed_pkt_dump = new PacketDumper(dumper);
		// If both -A and -w are specified, -A will be the transformed
		// trace file and -w will be the source packet trace file.
		// Otherwise the packets will go to the same file.
		if ( pkt_dump )
			source_pkt_dump = new PacketDumper(pkt_dump);
		}

	else if ( pkt_dump )
		transformed_pkt_dump = new PacketDumper(pkt_dump);

	if ( anonymize_ip_addr )
		init_ip_addr_anonymizers();
	else
		for ( int i = 0; i < NUM_ADDR_ANONYMIZATION_METHODS; ++i )
			ip_anonymizer[i] = 0;

	packet_sorter = new PacketSortGlobalPQ();

	sessions = new NetSessions();

	if ( do_watchdog )
		{
		// Set up the watchdog to make sure we don't wedge.
		(void) setsignal(SIGALRM, watchdog);
		(void) alarm(watchdog_interval);
		}
	}

void extract_next_packet(PktSrc* soonest_ps, double soonest,
			 struct pcap_pkthdr*& hdr,
			 const u_char*& pkt,
			 int& hdr_size,
			 double& ts)
	{
	current_pktsrc = soonest_ps;
	pkt = soonest_ps->NextPacket(hdr);
	hdr_size = soonest_ps->HdrSize();

	if ( encap_hdr_size > 0 )
		{
		// We're doing tunnel encapsulation.  Check whether there's
		// a particular associated port.
		if ( tunnel_port > 0 )
			{
			struct ip* ip_hdr = (struct ip*) (pkt + hdr_size);
			if ( ip_hdr->ip_p == IPPROTO_UDP )
				{
				struct udphdr* udp_hdr = (struct udphdr*)
					(pkt + hdr_size + ip_hdr->ip_hl * 4);

				if ( ntohs(udp_hdr->uh_dport) == tunnel_port )
					// A match.
					hdr_size += encap_hdr_size;
				}
			}

		else
			// Blanket encapsulation.
			hdr_size += encap_hdr_size;
		}

	ts = soonest;

	if ( ! pkt )
		internal_error("dispatch packet missing");

	soonest_ps->ConsumePacket();

	next_packet_time = soonest;
	}

int dispatch_next_packet()
	{
	int has_packet_to_dispatch = 0;
	const struct pcap_pkthdr* hdr = 0;
	const u_char* pkt;
	int hdr_size;
	double ts;
	PacketSortElement* packet_to_dispatch = 0;
	const PktSrc* src_ps;

	// Extract packets from sources till a packet can be
	// dispatched or all packets are exhausted.
	do {
		// Look for the packet source with the earliest timestamp
		// and dispatch it.  We assume here that there are few
		// sources, so that brute force across them is likely as or
		// more efficient than a fancier approach such as a priority
		// queue.
		double soonest = 0.0;
		PktSrc* soonest_ps = 0;

		loop_over_list(pkt_srcs, i)
			{
			PktSrc* ps = pkt_srcs[i];
			double t = ps->NextTimestamp();

			if ( t < 0.0 )
				{
				// Source has gone dry.  If it's a network
				// interface, this just means it's timed out.
				// If it's a file, though, then the file
				// has been exhausted, and we should delete
				// it as a packet source.
				if ( ! ps->IsLive() )
					{
					delete ps;
					pkt_srcs.remove_nth(i);
					--i;	// because it's about to be ++
					}
				continue;
				}

			if ( ! soonest_ps || t < soonest )
				{
				soonest_ps = ps;
				soonest = t;
				}
			}

		// Check if there are events received earlier from an
		// external source than our soonest packet. If so,
		// queue it and advance the time.
		if ( external_event_sources.length() > 0 )
			{
			EventSource* soonest_external_src = 0;
			int remove_src = -1;

			loop_over_list(external_event_sources, i)
				{
				EventSource* s = external_event_sources[i];
				if ( s->NextTimestamp() &&
				     (! soonest || s->NextTimestamp() < soonest) )
					{
					soonest_external_src = s;
					soonest = s->NextTimestamp();
					}
				else
					{
					// For simplicity, we remove at most
					// one inactive src.  (We cannot modify
					// the list during iteration.)
					if ( ! s->IsActive() )
						remove_src = i;
					}
				}

			if ( remove_src >= 0 )
				external_event_sources.remove_nth(remove_src);

			if ( soonest_external_src )
				{
				network_time = soonest;
				soonest_external_src->DispatchNextEvent();
				current_dispatched =
					timer_mgr->Advance(network_time,
							   max_timer_expires);
				return 1;
				}
			}

		if ( soonest_ps )
			next_packet_time = soonest;

		else if ( pkt_srcs.length() == 0 )
			next_packet_time = max_packet_time;

		src_ps = soonest_ps;

		if ( sort_packets() )
			{
			double min_t = next_packet_time - packet_sort_window;

			packet_to_dispatch = packet_sorter->RemoveMin(min_t);
			if ( packet_to_dispatch )
				{
				has_packet_to_dispatch = 1;
				src_ps = packet_to_dispatch->Src();
				hdr = packet_to_dispatch->Hdr();
				pkt = packet_to_dispatch->Pkt();
				hdr_size = packet_to_dispatch->HdrSize();
				ts = packet_to_dispatch->TimeStamp();
				break;
				}
			}

		else	// skip packet sort
			{
			if ( soonest_ps )
				{
				has_packet_to_dispatch = 1;
				struct pcap_pkthdr* next_hdr;
				extract_next_packet(soonest_ps, soonest, next_hdr, pkt, hdr_size, ts);
				hdr = next_hdr;
				}
			}

		if ( ! has_packet_to_dispatch )
			{
			if ( soonest_ps )
				{
				struct pcap_pkthdr* next_hdr;
				extract_next_packet(soonest_ps, soonest, next_hdr, pkt, hdr_size, ts);
				ASSERT(sort_packets());
				packet_sorter->Add(
					new PacketSortElement(soonest_ps, soonest, next_hdr, pkt, hdr_size));
				}
			else
				{
				if ( pkt_srcs.length() == 0 &&
				     packet_sorter->Empty() )
					// We've run out of sources.
					return remote_serializer &&
					       remote_serializer->IsActive();

				ASSERT(reading_live);
				if ( reading_live )
					{
					// Take advantage of the lull to get up
					// to date on timers and events.
					network_time = next_packet_time =
						current_time();
					timer_mgr->Advance(network_time,
							   max_timer_expires);
					mgr.Drain();
					return 1;
					}
				}
			}
	} while ( ! has_packet_to_dispatch );

	// Now dispatch the packet.
	if ( has_packet_to_dispatch )
		{
		current_hdr = hdr;
		current_pkt = pkt;

		processing_start_time = network_time = ts;

		{
		SegmentProfiler(segment_logger, "expiring-timers");
		current_dispatched =
			timer_mgr->Advance(network_time, max_timer_expires);
		}

		if ( src_ps->FilterType() == TYPE_FILTER_NORMAL )
			sessions->NextPacket(network_time, hdr, pkt, hdr_size,
						packet_to_dispatch);
		else
			sessions->NextPacketSecondary(network_time, hdr,
					pkt, hdr_size,
					packet_to_dispatch, src_ps);

		delete packet_to_dispatch;

		current_hdr = 0;	// done with these
		current_pkt = 0;

		mgr.Drain();

		extern void termination_signal();
		if ( signal_val == SIGTERM || signal_val == SIGINT )
			// We received a signal while processing the
			// current packet and its related events.
			termination_signal();

		processing_start_time = 0.0;	// = "we're not processing now"
		current_dispatched = 0;
		}

	if ( signal_val )
		{
		if ( bro_signal )
			{
			val_list* args = new val_list;
			args->append(new Val(signal_val, TYPE_COUNT));
			signal_val = 0;

			// Don't queue the event - deliver it
			// immediately.
			bro_signal->Call(args);
			delete args;
			}

		else
			{
			Val sval(signal_val, TYPE_COUNT);
			run_time("exiting due to signal %s", &sval);
			return 0;
			}
		}

	return 1;
	}

void net_run()
	{
	while ( dispatch_next_packet() )
		{
		if ( remote_serializer && remote_serializer->IsActive() )
			remote_serializer->Poll(pkt_srcs.length() == 0);
		}

	// Get the final statistics now, and not when net_finish() is
	// called, since that might happen quite a bit in the future
	// due to expiring pending timers, and we don't want to ding
	// for any packets dropped beyond this point.
	net_get_final_stats();
	}

void net_get_final_stats()
	{
	loop_over_list(pkt_srcs, i)
		{
		PktSrc* ps = pkt_srcs[i];

		if ( ps->IsLive() )
			{
			struct PktSrc::Stats s;
			ps->Statistics(&s);
			fprintf(stderr, "%d packets received on interface %s, %d dropped\n",
					s.received, ps->Interface(), s.dropped);
			}
		}
	}

void net_finish(int drain_events)
	{
	if ( drain_events )
		{
		if ( sessions )
			sessions->Drain();

		mgr.Drain();

		if ( sessions )
			sessions->Done();
		}

	if ( pkt_dump )
		pcap_dump_close(pkt_dump);

#ifdef USE_UHASH
	DEBUG_MSG("uhash: %d/%d\n", hash_cnt_uhash, hash_cnt_all);
#endif
#ifdef DEBUG_BRO
	extern int num_packets_held, num_packets_cleaned;
	DEBUG_MSG("packets clean up: %d/%d\n", num_packets_cleaned, num_packets_held);
#endif
	}

void net_delete()
	{
	delete sessions;
	delete packet_sorter;

	// Can't put this in net_finish() because packets might be
	// dumped when connections are deleted.
	if ( transformed_pkt_dump )
		delete transformed_pkt_dump;

	for ( int i = 0; i < NUM_ADDR_ANONYMIZATION_METHODS; ++i )
		delete ip_anonymizer[i];
	}

bool net_dump_current_packet(const char* file)
	{
	return current_pktsrc ? current_pktsrc->DumpCurrentPacket(file) : false;
	}


// net_packet_match
//
// Description:
//  - Checks if a packet matches a filter. It just wraps up a call to
//    [pcap.h's] bpf_filter().
//
// Inputs:
//  - fp: a BPF-compiled filter
//  - pkt: a pointer to the packet
//  - len: the original packet length
//  - caplen: the captured packet length. This is pkt length
//
// Output:
//  - return: 1 if the packet matches the filter, 0 otherwise

int net_packet_match(BPF_Program* fp, const u_char* pkt,
		     u_int len, u_int caplen)
	{
	// NOTE: I don't like too much un-const'ing the pkt variable.
	return bpf_filter(fp->GetProgram()->bf_insns, (u_char*) pkt, len, caplen);
	}
